# Part of Spatial Math Toolbox for Python
# Copyright (c) 2000 Peter Corke
# MIT Licence, see details in top-level file: LICENCE

"""
Classes to abstract 3D pose and orientation using matrices in SE(3) and SO(3)

To use::

    from spatialmath.pose3d import *
    T = SE3.Rx(0.3)

    import spatialmath as sm
    T = sm.SE3.Rx(0.3)


 .. inheritance-diagram:: spatialmath.pose3d
    :top-classes: collections.UserList
    :parts: 1
    
.. image:: ../figs/pose-values.png
"""

# pylint: disable=invalid-name

import numpy as np

from spatialmath import base
from spatialmath.baseposematrix import BasePoseMatrix


# ============================== SO3 =====================================#


class SO3(BasePoseMatrix):  
    """
    SO(3) matrix class

    This subclass represents rotations in 3D space.  Internally it is a 3x3 
    orthogonal matrix belonging to the group SO(3).

 .. inheritance-diagram:: spatialmath.pose3d.SO3
    :top-classes: collections.UserList
    :parts: 1
    """

    def __init__(self, arg=None, *, check=True):
        """
        Construct new SO(3) object

        :rtype: SO3 instance

        There are multiple call signatures:

        - ``SO3()`` is an ``SO3`` instance with one value -- a 3x3 identity
          matrix which corresponds to a null rotation
        - ``SO3(R)`` is an ``SO3`` instance with with the value ``R`` which is a
          3x3 numpy array representing an SO(3) rotation matrix.  If ``check``
          is ``True`` check the matrix belongs to SO(3).
        - ``SO3([R1, R2, ... RN])`` is an ``SO3`` instance wwith ``N`` values
          given by the elements ``Ri`` each of which is a 3x3 NumPy array
          representing an SO(3) matrix. If ``check`` is ``True`` check the
          matrix belongs to SO(3).
        - ``SO3([X1, X2, ... XN])`` is an ``SO3`` instance with ``N`` values
          given by the elements ``Xi`` each of which is an SO3 or SE3 instance.

        :SymPy: supported
        """
        super().__init__()
        
        if isinstance(arg, SE3):
            self.data = [base.t2r(x) for x in arg.data]

        elif not super().arghandler(arg, check=check):
            raise ValueError('bad argument to constructor')

    @staticmethod
    def _identity():
        return np.eye(3)
    # ------------------------------------------------------------------------ #
    @property
    def shape(self):
        """
        Shape of the object's interal matrix representation

        :return: (3,3)
        :rtype: tuple

        Each value within the ``SO3`` instance is a NumPy array of this shape.
        """
        return (3, 3)

    @property
    def R(self):
        """
        SO(3) or SE(3) as rotation matrix

        :return: rotational component
        :rtype: numpy.ndarray, shape=(3,3)

        ``x.R`` is the rotation matrix component of ``x`` as an array with
        shape (3,3). If ``len(x) > 1``, return an array with shape=(N,3,3).

        .. warning:: The i'th rotation matrix is ``x[i,:,:]`` or simply 
            ``x[i]``. This is different to the MATLAB version where the i'th
            rotation matrix is ``x(:,:,i)``.

        Example:

        .. runblock:: pycon

            >>> from spatialmath import SO3
            >>> x = SO3.Rx(0.3)
            >>> x.R

        :SymPy: supported
        """
        if len(self) == 1:
            return self.A[:3, :3]
        else:
            return np.array([x[:3, :3] for x in self.A])

    @property
    def n(self):
        """
        Normal vector of SO(3) or SE(3)

        :return: normal vector
        :rtype: numpy.ndarray, shape=(3,)

        This is the first column of the rotation submatrix, sometimes called the
        *normal vector*.  It is parallel to the x-axis of the frame defined by
        this pose.
        """
        return self.A[:3, 0]

    @property
    def o(self):
        """
        Orientation vector of SO(3) or SE(3)

        :return: orientation vector
        :rtype: numpy.ndarray, shape=(3,)

        This is the second column of the rotation submatrix, sometimes called
        the *orientation vector*.  It is parallel to the y-axis of the frame
        defined by this pose.
        """
        return self.A[:3, 1]

    @property
    def a(self):
        """
        Approach vector of SO(3) or SE(3)

        :return: approach vector
        :rtype: numpy.ndarray, shape=(3,)

        This is the third column of the rotation submatrix, sometimes called the
        *approach vector*.  It is parallel to the z-axis of the frame defined by
        this pose.
        """
        return self.A[:3, 2]

    # ------------------------------------------------------------------------ #

    def inv(self):
        """
        Inverse of SO(3)

        :return: inverse
        :rtype: SO2 instance

        Efficiently compute the inverse of each of the SO(3) values taking into
        account the matrix structure.  For an SO(3) matrix the inverse is the
        transpose.
        """
        if len(self) == 1:
            return SO3(self.A.T, check=False)
        else:
            return SO3([x.T for x in self.A], check=False)

    def eul(self, unit='rad', flip=False):
        r"""
        SO(3) or SE(3) as Euler angles

        :param unit: angular units: 'rad' [default], or 'deg'
        :type unit: str
        :return: 3-vector of Euler angles
        :rtype: ndarray(3,), ndarray(n,3)

        ``x.eul`` is the Euler angle representation of the rotation.  Euler angles are
        a 3-vector :math:`(\phi, \theta, \psi)` which correspond to consecutive
        rotations about the Z, Y, Z axes respectively.

        If ``len(x)`` is:

        - 1, return an ndarray with shape=(3,)
        - N>1, return ndarray with shape=(3,N)

        :seealso: :func:`~spatialmath.pose3d.SE3.Eul`, :func:`~spatialmath.base.transforms3d.tr2eul`
        :SymPy: not supported
        """
        if len(self) == 1:
            return base.tr2eul(self.A, unit=unit, flip=flip)
        else:
            return np.array([base.tr2eul(x, unit=unit, flip=flip) for x in self.A])

    def rpy(self, unit='rad', order='zyx'):
        """
        SO(3) or SE(3) as roll-pitch-yaw angles

        :param order: angle sequence order, default to 'zyx'
        :type order: str
        :param unit: angular units: 'rad' [default], or 'deg'
        :type unit: str
        :return: 3-vector of roll-pitch-yaw angles
        :rtype: ndarray(3,), ndarray(n,3)

        ``x.rpy`` is the roll-pitch-yaw angle representation of the rotation.  The angles are
        a 3-vector :math:`(r, p, y)` which correspond to successive rotations about the axes
        specified by ``order``:

            - ``'zyx'`` [default], rotate by yaw about the z-axis, then by pitch about the new y-axis,
              then by roll about the new x-axis.  Convention for a mobile robot with x-axis forward
              and y-axis sideways.
            - ``'xyz'``, rotate by yaw about the x-axis, then by pitch about the new y-axis,
              then by roll about the new z-axis. Convention for a robot gripper with z-axis forward
              and y-axis between the gripper fingers.
            - ``'yxz'``, rotate by yaw about the y-axis, then by pitch about the new x-axis,
              then by roll about the new z-axis. Convention for a camera with z-axis parallel
              to the optic axis and x-axis parallel to the pixel rows.

        If `len(x)` is:

        - 1, return an ndarray with shape=(3,)
        - N>1, return ndarray with shape=(3,N)

        :seealso: :func:`~spatialmath.pose3d.SE3.RPY`, :func:`~spatialmath.base.transforms3d.tr2rpy`
        :SymPy: not supported
        """
        if len(self) == 1:
            return base.tr2rpy(self.A, unit=unit, order=order)
        else:
            return np.array([base.tr2rpy(x, unit=unit, order=order) for x in self.A])

    def angvec(self, unit='rad'):
        r"""
        SO(3) or SE(3) as angle and rotation vector

        :param unit: angular units: 'rad' [default], or 'deg'
        :type unit: str
        :param check: check that rotation matrix is valid
        :type check: bool
        :return: :math:`(\theta, {\bf v})`
        :rtype: float, numpy.ndarray, shape=(3,)

        ``q.angvec()`` is a tuple :math:`(\theta, v)` containing the rotation 
        angle and a rotation axis which is equivalent to the rotation of
        the unit quaternion ``q``.

        By default the angle is in radians but can be changed setting `unit='deg'`.

        .. notes::

            - If the input is SE(3) the translation component is ignored.

        Example:

        .. runblock:: pycon

            >>> from spatialmath import UnitQuaternion
            >>> UnitQuaternion.Rz(0.3).angvec()

        :seealso: :func:`~spatialmath.quaternion.AngVec`, :func:`~angvec2r`
        """
        return base.tr2angvec(self.R, unit=unit)

    # ------------------------------------------------------------------------ #

    @staticmethod
    def isvalid(x, check=True):
        """
        Test if matrix is valid SO(3)

        :param x: matrix to test
        :type x: numpy.ndarray
        :return: ``True`` if the matrix is a valid element of SO(3), ie. it is a 3x3
            orthonormal matrix with determinant of +1.
        :rtype: bool

        :seealso: :func:`~spatialmath.base.transform3d.isrot`
        """
        return base.isrot(x, check=True)

    # ---------------- variant constructors ---------------------------------- #

    @classmethod
    def Rx(cls, theta, unit='rad'):
        """
        Construct a new SO(3) from X-axis rotation

        :param θ: rotation angle about the X-axis
        :type θ: float or array_like
        :param unit: angular units: 'rad' [default], or 'deg'
        :type unit: str
        :return: SO(3) rotation
        :rtype: SO3 instance

        - ``SE3.Rx(θ)`` is an SO(3) rotation of ``θ`` radians about the x-axis
        - ``SE3.Rx(θ, "deg")`` as above but ``θ`` is in degrees

        If ``theta`` is an array then the result is a sequence of rotations defined by consecutive
        elements.

        Example:

        .. runblock:: pycon

            >>> from spatialmath import SO3
            >>> x = SO3.Rx(np.linspace(0, math.pi, 20))
            >>> len(x)
            >>> x[7]

        """
        return cls([base.rotx(x, unit=unit) for x in base.getvector(theta)], check=False)

    @classmethod
    def Ry(cls, theta, unit='rad'):
        """
        Construct a new SO(3) from Y-axis rotation

        :param θ: rotation angle about Y-axis
        :type θ: float or array_like
        :param unit: angular units: 'rad' [default], or 'deg'
        :type unit: str
        :return: SO(3) rotation
        :rtype: SO3 instance

        - ``SO3.Ry(θ)`` is an SO(3) rotation of ``θ`` radians about the y-axis
        - ``SO3.Ry(θ, "deg")`` as above but ``θ`` is in degrees

        If ``θ`` is an array then the result is a sequence of rotations defined by consecutive
        elements.

        Example:

        .. runblock:: pycon

            >>> from spatialmath import UnitQuaternion
            >>> x = SO3.Ry(np.linspace(0, math.pi, 20))
            >>> len(x)
            >>> x[7]

        """
        return cls([base.roty(x, unit=unit) for x in base.getvector(theta)], check=False)

    @classmethod
    def Rz(cls, theta, unit='rad'):
        """
        Construct a new SO(3) from Z-axis rotation

        :param θ: rotation angle about Z-axis
        :type θ: float or array_like
        :param unit: angular units: 'rad' [default], or 'deg'
        :type unit: str
        :return: SO(3) rotation
        :rtype: SO3 instance

        - ``SO3.Rz(θ)`` is an SO(3) rotation of ``θ`` radians about the z-axis
        - ``SO3.Rz(θ, "deg")`` as above but ``θ`` is in degrees

        If ``θ`` is an array then the result is a sequence of rotations defined by consecutive
        elements.

        Example:

        .. runblock:: pycon

            >>> from spatialmath import SE3
            >>> x = SE3.Rz(np.linspace(0, math.pi, 20))
            >>> len(x)
            >>> x[7]

        """
        return cls([base.rotz(x, unit=unit) for x in base.getvector(theta)], check=False)

    @classmethod
    def Rand(cls, N=1):
        """
        Construct a new SO(3) from random rotation

        :param N: number of random rotations
        :type N: int
        :return: SO(3) rotation matrix
        :rtype: SO3 instance

        - ``SO3.Rand()`` is a random SO(3) rotation.
        - ``SO3.Rand(N)`` is a sequence of N random rotations.

        Example:

        .. runblock:: pycon

            >>> from spatialmath import SO3
            >>> x = SO3.Rand()
            >>> x

        :seealso: :func:`spatialmath.quaternion.UnitQuaternion.Rand`
        """
        return cls([base.q2r(base.qrand()) for _ in range(0, N)], check=False)

    @classmethod
    def Eul(cls, *angles, unit='rad'):
        r"""
        Construct a new SO(3) from Euler angles

        :param 𝚪: Euler angles
        :type 𝚪: array_like or numpy.ndarray with shape=(N,3)
        :param unit: angular units: 'rad' [default], or 'deg'
        :type unit: str
        :return: SO(3) rotation
        :rtype: SO3 instance

        ``SO3.Eul(𝚪)`` is an SO(3) rotation defined by a 3-vector of Euler
          angles :math:`\Gamma = (\phi, \theta, \psi)` which correspond to
          consecutive rotations about the Z, Y, Z axes respectively. If ``𝚪``
          is an Nx3 matrix then the result is a sequence of rotations each
          defined by Euler angles corresponding to the rows of ``angles``.

        ``SO3.Eul(φ, θ, ψ)`` as above but the angles are provided as three
          scalars.

        Example:

        .. runblock:: pycon
        
            >>> from spatialmath import SO3
            >>> SO3.Eul(0.1, 0.2, 0.3)
            >>> SO3.Eul([0.1, 0.2, 0.3])
            >>> SO3.Eul(10, 20, 30, 'deg')

        :seealso: :func:`~spatialmath.pose3d.SE3.eul`, :func:`~spatialmath.pose3d.SE3.Eul`, :func:`~spatialmath.base.transforms3d.eul2r`
        """
        if len(angles) == 1:
            angles = angles[0]

        if base.isvector(angles, 3):
            return cls(base.eul2r(angles, unit=unit), check=False)
        else:
            return cls([base.eul2r(a, unit=unit) for a in angles], check=False)

    @classmethod
    def RPY(cls, *angles, unit='rad', order='zyx', ):
        r"""
        Construct a new SO(3) from roll-pitch-yaw angles

        :param angles: roll-pitch-yaw angles
        :type angles: array_like(3), array_like(n,3)
        :param unit: angular units: 'rad' [default], or 'deg'
        :type unit: str
        :param order: rotation order: 'zyx' [default], 'xyz', or 'yxz'
        :type order: str
        :return: SO(3) rotation
        :rtype: SO3 instance

        - ``SO3.RPY(angles)`` is an SO(3) rotation defined by a 3-vector of
          roll, pitch, yaw angles :math:`(\alpha, \beta, \gamma)`. If ``angles``
          is an Nx3 matrix then the result is a sequence of rotations each
          defined by RPY angles corresponding to the rows of angles. The angles
          correspond to successive rotations about the axes specified by
          ``order``:

             - ``'zyx'`` [default], rotate by yaw about the z-axis, then by pitch about the new y-axis,
               then by roll about the new x-axis.  Convention for a mobile robot with x-axis forward
               and y-axis sideways.
            - ``'xyz'``, rotate by yaw about the x-axis, then by pitch about the new y-axis,
              then by roll about the new z-axis. Convention for a robot gripper with z-axis forward
              and y-axis between the gripper fingers.
            - ``'yxz'``, rotate by yaw about the y-axis, then by pitch about the new x-axis,
              then by roll about the new z-axis. Convention for a camera with z-axis parallel
              to the optic axis and x-axis parallel to the pixel rows.

        - ``SO3.RPY(⍺, β, 𝛾)`` as above but the angles are provided as three
          scalars.

        Example:

        .. runblock:: pycon
        
            >>> from spatialmath import SO3
            >>> SO3.RPY(0.1, 0.2, 0.3)
            >>> SO3.RPY([0.1, 0.2, 0.3])
            >>> SO3.RPY(0.1, 0.2, 0.3, order='xyz')
            >>> SO3.RPY(10, 20, 30, 'deg')


        :seealso: :func:`~spatialmath.pose3d.SE3.rpy`, :func:`~spatialmath.pose3d.SE3.RPY`, :func:`spatialmath.base.transforms3d.rpy2r`
        """
        if len(angles) == 1:
            angles = angles[0]

        # angles = base.getmatrix(angles, (None, 3))
        # return cls(base.rpy2r(angles, order=order, unit=unit), check=False)

        if base.isvector(angles, 3):
            return cls(base.rpy2r(angles, unit=unit, order=order), check=False)
        else:
            return cls([base.rpy2r(a, unit=unit, order=order) for a in angles], check=False)

    @classmethod
    def OA(cls, o, a):
        """
        Construct a new SO(3) from two vectors

        :param o: 3-vector parallel to Y- axis
        :type o: array_like
        :param a: 3-vector parallel to the Z-axis
        :type o: array_like
        :return: SO(3) rotation
        :rtype: SO3 instance

        ``SO3.OA(O, A)`` is an SO(3) rotation defined in terms of
        vectors parallel to the Y- and Z-axes of its reference frame.  In robotics these axes are
        respectively called the *orientation* and *approach* vectors defined such that
        R = [N, O, A] and N = O x A.

        .. notes::

            - Only the ``A`` vector is guaranteed to have the same direction in the resulting
            rotation matrix
            - ``O`` and ``A`` do not have to be unit-length, they are normalized
            - ``O`` and ``A` do not have to be orthogonal, so long as they are not parallel

        :seealso: :func:`spatialmath.base.transforms3d.oa2r`
        """
        return cls(base.oa2r(o, a), check=False)

    @classmethod
    def TwoVectors(cls, x=None, y=None, z=None):
        """
        Construct a new SO(3) from any two vectors

        :param x: new x-axis, defaults to None
        :type x: str, array_like(3), optional
        :param y: new y-axis, defaults to None
        :type y: str, array_like(3), optional
        :param z: new z-axis, defaults to None
        :type z: str, array_like(3), optional

        Create a rotation by defining the direction of two of the new
        axes in terms of the old axes.  Axes are denoted by strings ``"x"``, 
        ``"y"``, ``"z"``, ``"-x"``, ``"-y"``, ``"-z"``.

        The directions can also be specified by 3-element vectors, but these
        must be orthogonal.

        To create a rotation where the new frame has its x-axis in -z-direction
        of the previous frame, and its z-axis in the x-direction of the previous
        frame is::
        
            >>> SO3.TwoVectors(x='-z', z='x')
        """
        def vval(v):
            if isinstance(v, str):
                sign = 1
                if v[0] == '-':
                    sign = -1
                    v = v[1:] # skip sign char
                elif v[0] == '+':
                    v = v[1:] # skip sign char
                if v[0] == 'x':
                    v = [sign, 0, 0]
                elif v[0] == 'y':
                    v = [0, sign, 0]
                elif v[0] == 'z':
                    v = [0, 0, sign]
                return np.r_[v]
            else:
                return base.unitvec(base.getvector(v, 3))

        if x is not None and y is not None and z is None:
            # z = x x y
            x = vval(x)
            y = vval(y)
            z = np.cross(x, y)

        elif x is None and y is not None and z is not None:
            # x = y x z
            y = vval(y)
            z = vval(z)
            x = np.cross(y, z)

        elif x is not None and y is None and z is not None:
            # y = z x x
            z = vval(z)
            x = vval(x)
            y = np.cross(z, x)

        return cls(np.c_[x, y, z], check=False)

    @classmethod
    def AngleAxis(cls, theta, v, *, unit='rad'):
        r"""
        Construct a new SO(3) rotation matrix from rotation angle and axis

        :param theta: rotation
        :type theta: float
        :param unit: angular units: 'rad' [default], or 'deg'
        :type unit: str
        :param v: rotation axis, 3-vector
        :type v: array_like
        :return: SO(3) rotation
        :rtype: SO3 instance

        ``SO3.AngleAxis(theta, V)`` is an SO(3) rotation defined by
        a rotation of ``THETA`` about the vector ``V``.

        .. note:: :math:`\theta \eq 0` the result in an identity matrix, otherwise
            ``V`` must have a finite length, ie. :math:`|V| > 0`.

        :seealso: :func:`~spatialmath.pose3d.SE3.angvec`, :func:`spatialmath.base.transforms3d.angvec2r`
        """
        return cls(base.angvec2r(theta, v, unit=unit), check=False)
        
    @classmethod
    def AngVec(cls, theta, v, *, unit='rad'):
        r"""
        Construct a new SO(3) rotation matrix from rotation angle and axis

        :param theta: rotation
        :type theta: float
        :param unit: angular units: 'rad' [default], or 'deg'
        :type unit: str
        :param v: rotation axis, 3-vector
        :type v: array_like
        :return: SO(3) rotation
        :rtype: SO3 instance

        ``SO3.AngVec(theta, V)`` is an SO(3) rotation defined by
        a rotation of ``THETA`` about the vector ``V``.

        .. deprecated:: 0.9.8
            Use :meth:`AngleAxis` instead.

        :seealso: :func:`~spatialmath.pose3d.SE3.angvec`, :func:`spatialmath.base.transforms3d.angvec2r`
        """
        return cls(base.angvec2r(theta, v, unit=unit), check=False)

    @classmethod
    def EulerVec(cls, w):
        r"""
        Construct a new SO(3) rotation matrix from an Euler rotation vector

        :param ω: rotation axis
        :type ω: 3-element array_like
        :return: SO(3) rotation
        :rtype: SO3 instance

        ``SO3.EulerVec(ω)`` is a unit quaternion that describes the 3D rotation
        defined by a rotation of :math:`\theta = \lVert \omega \rVert` about the
        unit 3-vector :math:`\omega / \lVert \omega \rVert`.

        Example:

        .. runblock:: pycon
        
            >>> from spatialmath import SO3
            >>> SO3.EulerVec([0.5,0,0])

        .. note:: :math:`\theta \eq 0` the result in an identity matrix, otherwise
            ``V`` must have a finite length, ie. :math:`|V| > 0`.

        :seealso: :func:`~spatialmath.pose3d.SE3.angvec`, :func:`~spatialmath.base.transforms3d.angvec2r`
        """
        assert base.isvector(w, 3), 'w must be a 3-vector'
        w = base.getvector(w)
        theta = base.norm(w)
        return cls(base.angvec2r(theta, w), check=False)

    @classmethod
    def Exp(cls, S, check=True, so3=True):
        r"""
        Create an SO(3) rotation matrix from so(3)

        :param S: Lie algebra so(3)
        :type S: numpy ndarray
        :param check: check that passed matrix is valid so(3), default True
        :type check: bool
        :return: SO(3) rotation
        :rtype: SO3 instance

        - ``SO3.Exp(S)`` is an SO(3) rotation defined by its Lie algebra
          which is a 3x3 so(3) matrix (skew symmetric)
        - ``SO3.Exp(t)`` is an SO(3) rotation defined by a 3-element twist
          vector (the unique elements of the so(3) skew-symmetric matrix)
        - ``SO3.Exp(T)`` is a sequence of SO(3) rotations defined by an Nx3 matrix
          of twist vectors, one per row.

        Note:
        - if :math:`\theta \eq 0` the result in an identity matrix
        - an input 3x3 matrix is ambiguous, it could be the first or third case above.  In this
          case the parameter `so3` is the decider.

        :seealso: :func:`spatialmath.base.transforms3d.trexp`, :func:`spatialmath.base.transformsNd.skew`
        """
        if base.ismatrix(S, (-1, 3)) and not so3:
            return cls([base.trexp(s, check=check) for s in S], check=False)
        else:
            return cls(base.trexp(S, check=check), check=False)

    def angdist(self, other, metric=6):
        r"""
        Angular distance metric between rotations

        :param other: second rotation
        :type other: SO3 instance
        :param metric: metric, default is 6
        :type metric: int
        :raises TypeError: if other is not an SO3
        :return: angle in radians
        :rtype: float or ndarray

        ``R1.angdist(R2)`` is the geodesic norm, or geodesic distance between two
        rotations.

        Several metrics are supported, the first 5 are computed after conversion
        to unit quaternions.

        ======   ===============================================================
        Metric   Details
        ======   ===============================================================
        0        :math:`1 - | \q_1 \bullet \q_2 | \in [0, 1]`
        1        :math:`\cos^{-1} | \q_1 \bullet \q_2 | \in [0, \pi/2]`
        2        :math:`\cos^{-1} | \q_1 \bullet \q_2 | \in [0, \pi/2]`
        3        :math:`2 \tan^{-1} \| \q_1 - \q_2\| / \|\q_1 + \q_2\| \in [0, \pi/2]`
        4        :math:`\cos^{-1} \left( 2 (\q_1 \bullet \q_2)^2 - 1\right) \in [0, 1]`
        5        :math:`\|I - \mat{R}_1 \mat{R}_2^T\| \in [0, 2]`
        6        :math:`\|\log \mat{R}_1 \mat{R}_2^T\| \in [0, \pi]`
        ======   ===============================================================

        Example:

        .. runblock:: pycon

            >>> from spatialmath import SO3
            >>> R1 = SO3.Rx(0.3)
            >>> R2 = SO3.Ry(0.3)
            >>> print(R1.angdist(R1))
            >>> print(R1.angdist(R2))

        .. note::
            - metrics 1, 2, 4 can throw ValueError "math domain error" due to
              numeric errors which push the argument of ``acos()`` marginally
              outside its domain [0, 1].
            - metrics 2 and 3 are equivalent, but 3 is more robust

        :seealso: :func:`UnitQuaternion.angdist`
        """

        if metric < 5:
            from spatialmath.quaternion import UnitQuaternion

            return UnitQuaternion(self).angdist(UnitQuaternion(other), metric=metric)

        elif metric == 5:
            op = lambda R1, R2: np.linalg.norm(np.eye(3) - R1 @ R2.T)
        elif metric == 6:
            op = lambda R1, R2: base.norm(base.trlog(R1 @ R2.T, twist=True))
        else:
            raise ValueError('unknown metric')
        
        ad = self._op2(other, op)
        if isinstance(ad, list):
            return np.array(ad)
        else:
            return ad

# ============================== SE3 =====================================#


class SE3(SO3):
    """
    SE(3) matrix class

    This subclass represents rigid-body motion in 3D space.  Internally it is a 
    4x4 homogeneous transformation matrix belonging to the group SE(3).

 .. inheritance-diagram:: spatialmath.pose3d.SE3
    :top-classes: collections.UserList
    :parts: 1
    """

    def __init__(self, x=None, y=None, z=None, *, check=True):
        """
        Construct new SE(3) object

        :rtype: SE3 instance

        There are multiple call signatures that return an ``SE3`` instance
        with one or more values.

        - ``SE3()`` null motion, value is the identity matrix.
        - ``SE3(x, y, z)`` is a pure translation of (x,y,z)
        - ``SE3(T)``  where ``T`` is a 4x4 Numpy  array representing an SE(3)
          matrix.  If ``check`` is ``True`` check the matrix belongs to SE(3).
        - ``SE3([T1, T2, ... TN])`` has ``N`` values
          given by the elements ``Ti`` each of which is a 4x4 NumPy array
          representing an SE(3) matrix. If ``check`` is ``True`` check the
          matrix belongs to SE(3).
        - ``SE3(X)`` where ``X`` is:
          -  ``SE3`` is a copy of ``X``
          -  ``SO3`` is the rotation of ``X`` with zero translation
          -  ``SE2`` is the z-axis rotation and x- and y-axis translation of
             ``X``
        - ``SE3([X1, X2, ... XN])`` has ``N`` values
          given by the elements ``Xi`` each of which is an SE3 instance.
        
        :SymPy: supported
        """
        if y is None and z is None:
            # just one argument passed

            if super().arghandler(x, check=check):
                return
            elif isinstance(x, SO3):
                self.data = [base.r2t(_x) for _x in x.data]
            elif type(x).__name__ == 'SE2':
                def convert(x):
                    # convert SE(2) to SE(3)
                    out = np.identity(4, dtype=x.dtype)
                    out[:2,:2] = x[:2,:2]
                    out[:2,3] = x[:2,2]
                    return out
                self.data = [convert(_x) for _x in x.data]
            elif base.isvector(x, 3):
                # SE3( [x, y, z] )
                self.data = [base.transl(x)]
            elif isinstance(x, np.ndarray) and x.shape[1] == 3:
                # SE3( Nx3 )
                self.data = [base.transl(T) for T in x]

            else:
                raise ValueError('bad argument to constructor')

        elif y is not None and z is not None:
            # SE3(x, y, z)
            self.data = [base.transl(x, y, z)]

    @staticmethod
    def _identity():
        return np.eye(4)
        
    # ------------------------------------------------------------------------ #
    @property
    def shape(self):
        """
        Shape of the object's internal matrix representation

        :return: (4,4)
        :rtype: tuple

        Each value within the ``SE3`` instance is a NumPy array of this shape.
        """
        return (4, 4)

    @property
    def t(self):
        """
        Translational component of SE(3)

        :return: translational component of SE(3)
        :rtype: numpy.ndarray

        ``x.t`` is the translational component of ``x`` as an array with
        shape (3,). If ``len(x) > 1``, return an array with shape=(N,3).

        Example:

        .. runblock:: pycon

            >>> from spatialmath import SE3
            >>> x = SE3(1,2,3)
            >>> x.t
            >>> x = SE3([ SE3(1,2,3), SE3(4,5,6)])
            >>> x.t
        
        :SymPy: supported
        """
        if len(self) == 1:
            return self.A[:3, 3]
        else:
            return np.array([x[:3, 3] for x in self.A])

    @t.setter
    def t(self, v):
        if len(self) > 1:
            raise ValueError("can only assign translation to length 1 object")
        v = base.getvector(v, 3)
        self.A[:3, 3] = v
    # ------------------------------------------------------------------------ #

    def inv(self):
        r"""
        Inverse of SE(3)

        :return: inverse
        :rtype: SE3 instance

        Efficiently compute the inverse of each of the SE(3) values taking into
        account the matrix structure.

        .. math::
        
            T = \left[ \begin{array}{cc} \mat{R} & \vec{t} \\ 0 & 1 \end{array} \right],
            \mat{T}^{-1} = \left[ \begin{array}{cc} \mat{R}^T & -\mat{R}^T \vec{t} \\ 0 & 1 \end{array} \right]`

        Example:

        .. runblock:: pycon

            >>> from spatialmath import SE3
            >>> x = SE3(1,2,3)
            >>> x.inv()


        :seealso: :func:`~spatialmath.base.transforms3d.trinv`

        :SymPy: supported
        """
        if len(self) == 1:
            return SE3(base.trinv(self.A), check=False)
        else:
            return SE3([base.trinv(x) for x in self.A], check=False)

    def delta(self, X2=None):
        r"""
        Infinitesimal difference of SE(3) values

        :return: differential motion vector
        :rtype: numpy.ndarray, shape=(6,)

        ``X1.delta(X2)`` is the differential motion (6x1) corresponding to
        infinitesimal motion (in the ``X1`` frame) from pose ``X1`` to ``X2``.

        The vector :math:`d = [\delta_x, \delta_y, \delta_z, \theta_x, \theta_y, \theta_z]`
        represents infinitesimal translation and rotation.

        Example:

        .. runblock:: pycon

            >>> from spatialmath import SE3
            >>> x1 = SE3.Rx(0.3)
            >>> x2 = SE3.Rx(0.3001)
            >>> x1.delta(x2)


        .. note::

            - the displacement is only an approximation to the motion, and assumes
              that ``X1`` ~ ``X2``.
            - can be considered as an approximation to the effect of spatial velocity over a
              a time interval, ie. the average spatial velocity multiplied by time.

        :Reference: Robotics, Vision & Control: Second Edition, P. Corke, Springer 2016; p67.

        :seealso: :func:`~spatialmath.base.transforms3d.tr2delta`
        """
        if X2 is None:
            return base.tr2delta(self.A)
        else:
            return base.tr2delta(self.A, X2.A)

    def Ad(self):
        r"""
        Adjoint of SE(3)

        :return: adjoint matrix
        :rtype: numpy.ndarray, shape=(6,6)

        ``SE3.Ad`` is the 6x6 adjoint matrix

        If spatial velocity :math:`\nu = (v_x, v_y, v_z, \omega_x, \omega_y, \omega_z)^T`
        and the SE(3) represents the pose of {B} relative to {A}, 
        ie. :math:`{}^A {\bf T}_B, and the adjoint is :math:`\mathbf{A}` then
        :math:`{}^{A}\!\nu = \mathbf{A} {}^{B}\!\nu`.

        .. warning:: Do not use this method to map velocities 
            between robot base and end-effector frames - use ``jacob()``.

        .. note:: Use this method to map velocities between two frames on
            the same rigid-body.  

        :reference: Robotics, Vision & Control: Second Edition, P. Corke, Springer 2016; p65.
        :seealso: SE3.jacob, Twist.ad, :func:`~spatialmath.base.tr2jac`
        :SymPy: supported
        """
        return base.tr2adjoint(self.A)

    def jacob(self):
        r"""
        Velocity transform for SE(3)

        :return: Jacobian matrix
        :rtype: numpy.ndarray, shape=(6,6)

        ``SE3.jacob()`` is the 6x6 Jacobian that maps spatial velocity or
        differential motion from frame {B} to frame {A} where the pose of {B}
        relative to {A} is represented by the homogeneous transform T =
        :math:`{}^A {\bf T}_B`.  
        
        .. note::
            - To map from frame {A} to frame {B} use the transpose of this matrix.
            - Use this method to map velocities between the robot end-effector frame
              and the base frames.

        .. warning:: Do not use this method to map velocities between two frames
            on the same rigid-body.

        :seealso: SE3.Ad, Twist.ad, :func:`~spatialmath.base.tr2jac`
        :Reference: Robotics, Vision & Control: Second Edition, P. Corke, Springer 2016; p65.
        :SymPy: supported
        """
        return base.tr2jac(self.A)

    def twist(self):
        """
        SE(3) as twist

        :return: equivalent rigid-body motion as a twist vector
        :rtype: Twist3 instance

        Example:

        .. runblock:: pycon

            >>> from spatialmath import SE3
            >>> x = SE3(1,2,3)
            >>> x.twist()

        :seealso: :func:`spatialmath.twist.Twist3`
        """
        from spatialmath.twist import Twist3

        return Twist3(self.log(twist=True))
    # ------------------------------------------------------------------------ #

    @staticmethod
    def isvalid(x, check=True):
        """
        Test if matrix is a valid SE(3)

        :param x: matrix to test
        :type x: numpy.ndarray
        :return: ``True`` if the matrix is 4x4 and a valid element of SE(3), ie. it
                 is a valid homogeneous transformation matrix.
        :rtype: bool

        :seealso: :func:`~spatialmath.base.transforms3d.ishom`
        """
        return base.ishom(x, check=check)

    # ---------------- variant constructors ---------------------------------- #

    @classmethod
    def Rx(cls, theta, unit='rad', t=None):
        """
        Create anSE(3) pure rotation about the X-axis

        :param θ: rotation angle about X-axis
        :type θ: float
        :param unit: angular units: 'rad' [default], or 'deg'
        :type unit: str
        :param t: translation, optional
        :type t: 3-element array-like
        :return: SE(3) matrix
        :rtype: SE3 instance

        - ``SE3.Rx(θ)`` is an SE(3) rotation of θ radians about the x-axis
        - ``SE3.Rx(θ, "deg")`` as above but θ is in degrees
        - ``SE3.Rx(θ, t=T)`` as above but also sets the translational component

        If ``θ`` is an array then the result is a sequence of rotations defined
        by consecutive elements.

        .. note:: The translation option only works for the scalar θ case.

        Example:

        .. runblock:: pycon

            >>> from spatialmath import SE3
            >>> SE3.Rx(0.3)
            >>> SE3.Rx([0.3, 0.4])

        :seealso: :func:`~spatialmath.base.transforms3d.trotx`
        :SymPy: supported
        """
        return cls([base.trotx(x, t=t, unit=unit) for x in base.getvector(theta)], check=False)

    @classmethod
    def Ry(cls, theta, unit='rad', t=None):
        """
        Create an SE(3) pure rotation about the Y-axis

        :param θ: rotation angle about X-axis
        :type θ: float
        :param unit: angular units: 'rad' [default], or 'deg'
        :type unit: str
        :param t: translation, optional
        :type t: 3-element array-like
        :return: SE(3) matrix
        :rtype: SE3 instance

        - ``SE3.Ry(θ)`` is an SO(3) rotation of θ radians about the y-axis
        - ``SE3.Ry(θ, "deg")`` as above but θ is in degrees
        - ``SE3.Ry(θ, t=T)`` as above but also sets the translational component

        If ``θ`` is an array then the result is a sequence of rotations defined
        by consecutive elements.

        .. note:: The translation option only works for the scalar θ case.

        Example:

        .. runblock:: pycon

            >>> from spatialmath import SE3
            >>> SE3.Ry(0.3)
            >>> SE3.Ry([0.3, 0.4])

        :seealso: :func:`~spatialmath.base.transforms3d.troty`
        :SymPy: supported
        """
        return cls([base.troty(x, t=t, unit=unit) for x in base.getvector(theta)], check=False)

    @classmethod
    def Rz(cls, theta, unit='rad', t=None):
        """
        Create an SE(3) pure rotation about the Z-axis

        :param θ: rotation angle about Z-axis
        :type θ: float
        :param unit: angular units: 'rad' [default], or 'deg'
        :type unit: str
        :param t: translation, optional
        :type t: 3-element array-like
        :return: SE(3) matrix
        :rtype: SE3 instance

        - ``SE3.Rz(θ)`` is an SO(3) rotation of θ radians about the z-axis
        - ``SE3.Rz(θ, "deg")`` as above but θ is in degrees
        - ``SE3.Rz(θ, t=T)`` as above but also sets the translational component

        If ``θ`` is an array then the result is a sequence of rotations defined
        by consecutive elements.

        .. note:: The translation option only works for the scalar θ case.

        Example:

        .. runblock:: pycon

            >>> from spatialmath import SE3
            >>> SE3.Rz(0.3)
            >>> SE3.Rz([0.3, 0.4])

        :seealso: :func:`~spatialmath.base.transforms3d.trotz`
        :SymPy: supported
        """
        return cls([base.trotz(x, t=t, unit=unit) for x in base.getvector(theta)], check=False)

    @classmethod
    def Rand(cls, N=1, xrange=(-1, 1), yrange=(-1, 1), zrange=(-1, 1)):  # pylint: disable=arguments-differ
        """
        Create a random SE(3)

        :param xrange: x-axis range [min,max], defaults to [-1, 1]
        :type xrange: 2-element sequence, optional
        :param yrange: y-axis range [min,max], defaults to [-1, 1]
        :type yrange: 2-element sequence, optional
        :param zrange: z-axis range [min,max], defaults to [-1, 1]
        :type zrange: 2-element sequence, optional
        :param N: number of random transforms
        :type N: int
        :return: SE(3) matrix
        :rtype: SE3 instance

        Return an SE3 instance with random rotation and translation.

        - ``SE3.Rand()`` is a random SE(3) translation.
        - ``SE3.Rand(N)`` is an SE3 object containing a sequence of N random
          poses.


        Example:

        .. runblock:: pycon

            >>> from spatialmath import SE3
            >>> SE3.Rand(2)

        :seealso: :func:`~spatialmath.quaternions.UnitQuaternion.Rand`
        """
        X = np.random.uniform(low=xrange[0], high=xrange[1], size=N)  # random values in the range
        Y = np.random.uniform(low=yrange[0], high=yrange[1], size=N)  # random values in the range
        Z = np.random.uniform(low=zrange[0], high=zrange[1], size=N)  # random values in the range
        R = SO3.Rand(N=N)
        return cls([base.transl(x, y, z) @ base.r2t(r.A) for (x, y, z, r) in zip(X, Y, Z, R)], check=False)

    @classmethod
    def Eul(cls, *angles, unit='rad'):
        r"""
        Create an SE(3) pure rotation from Euler angles

        :param 𝚪: Euler angles
        :type 𝚪: array_like or numpy.ndarray with shape=(N,3)
        :param unit: angular units: 'rad' [default], or 'deg'
        :type unit: str
        :return: SE(3) matrix
        :rtype: SE3 instance

        - ``SE3.Eul(𝚪)`` is an SE(3) rotation defined by a 3-vector of Euler
          angles :math:`\Gamma=(\phi, \theta, \psi)` which correspond to
          consecutive rotations about the Z, Y, Z axes respectively.

        If ``𝚪`` is an Nx3 matrix then the result is a sequence of
        rotations each defined by Euler angles corresponding to the rows of
        ``𝚪``.

        - ``SE3.Eul(φ, θ, ψ)`` as above but the angles are provided as three
          scalars.

        Example:

        .. runblock:: pycon
        
            >>> from spatialmath import SE3
            >>> SE3.Eul(0.1, 0.2, 0.3)
            >>> SE3.Eul([0.1, 0.2, 0.3])
            >>> SE3.Eul(10, 20, 30, unit='deg')

        :seealso: :func:`~spatialmath.pose3d.SE3.eul`, :func:`~spatialmath.base.transforms3d.eul2r`
        :SymPy: supported
        """
        if len(angles) == 1:
            angles = angles[0]
        if base.isvector(angles, 3):
            return cls(base.eul2tr(angles, unit=unit), check=False)
        else:
            return cls([base.eul2tr(a, unit=unit) for a in angles], check=False)

    @classmethod
    def RPY(cls, *angles, unit='rad', order='zyx'):
        r"""
        Create an SE(3) pure rotation from roll-pitch-yaw angles

        :param 𝚪: roll-pitch-yaw angles
        :type 𝚪: array_like or numpy.ndarray with shape=(N,3)
        :param unit: angular units: 'rad' [default], or 'deg'
        :type unit: str
        :param order: rotation order: 'zyx' [default], 'xyz', or 'yxz'
        :type order: str
        :return: SE(3) matrix
        :rtype: SE3 instance

        - ``SE3.RPY(𝚪)`` is an SE(3) rotation defined by a 3-vector of roll,
          pitch, yaw angles :math:`\Gamma=(r, p, y)` which correspond to
          successive rotations about the axes specified by ``order``:

            - ``'zyx'`` [default], rotate by yaw about the z-axis, then by pitch about the new y-axis,
              then by roll about the new x-axis.  This is the **convention** for a mobile robot with x-axis forward
              and y-axis sideways.
            - ``'xyz'``, rotate by yaw about the x-axis, then by pitch about the new y-axis,
              then by roll about the new z-axis. This is the **convention** for a robot gripper with z-axis forward
              and y-axis between the gripper fingers.
            - ``'yxz'``, rotate by yaw about the y-axis, then by pitch about the new x-axis,
              then by roll about the new z-axis. This is the **convention** for a camera with z-axis parallel
              to the optical axis and x-axis parallel to the pixel rows.

        If ``𝚪`` is an Nx3 matrix then the result is a sequence of rotations each defined by RPY angles
        corresponding to the rows of ``𝚪``.

        - ``SE3.RPY(⍺, β, 𝛾)`` as above but the angles are provided as three
          scalars.

        Example:

        .. runblock:: pycon
        
            >>> from spatialmath import SE3
            >>> SE3.RPY(0.1, 0.2, 0.3)
            >>> SE3.RPY([0.1, 0.2, 0.3])
            >>> SE3.RPY(0.1, 0.2, 0.3, order='xyz')
            >>> SE3.RPY(10, 20, 30, unit='deg')

        :seealso: :func:`~spatialmath.pose3d.SE3.rpy`, :func:`~spatialmath.base.transforms3d.rpy2r`
        :SymPy: supported
        """
        if len(angles) == 1:
            angles = angles[0]

        if base.isvector(angles, 3):
            return cls(base.rpy2tr(angles, order=order, unit=unit), check=False)
        else:
            return cls([base.rpy2tr(a, order=order, unit=unit) for a in angles], check=False)

    @classmethod
    def OA(cls, o, a):
        r"""
        Create an SE(3) pure rotation from two vectors

        :param o: 3-vector parallel to Y- axis
        :type o: array_like
        :param a: 3-vector parallel to the Z-axis
        :type a: array_like
        :return: SE(3) matrix
        :rtype: SE3 instance

        ``SE3.OA(o, a)`` is an SE(3) rotation defined in terms of vectors ``o``
        and ``a`` respectively parallel to the Y- and Z-axes of its reference
        frame.  In robotics these axes are respectively called the *orientation*
        and *approach* vectors defined such that :math:`\mathbf{R} = [n, o, a]`
        and :math:`n = o \times a`.

        .. note::

            - The ``a`` vector is the only guaranteed to have the same direction in the resulting
              rotation matrix
            - ``o`` and ``a`` do not have to be unit-length, they are normalized
            - ``o`` and ``a`` do not have to be orthogonal, so long as they are not parallel
              ``o`` is adjusted to be orthogonal to ``a``.

        Example:

        .. runblock:: pycon

            >>> from spatialmath import SE3
            >>> SE3.OA([1, 0, 0], [0, 0, -1])

        :seealso: :func:`~spatialmath.base.transforms3d.oa2r`
        """
        return cls(base.oa2tr(o, a), check=False)

    @classmethod
    def AngleAxis(cls, theta, v, *, unit='rad'):
        r"""
        Create an SE(3) pure rotation matrix from rotation angle and axis

        :param θ: rotation
        :type θ: float
        :param unit: angular units: 'rad' [default], or 'deg'
        :type unit: str
        :param v: rotation axis, 3-vector
        :type v: array_like
        :return: SE(3) matrix
        :rtype: SE3 instance

        ``SE3.AngleAxis(θ, v)`` is an SE(3) rotation defined by
        a rotation of ``θ`` about the vector ``v``.

        .. math::
        
            \mbox{if}\,\, \theta \left\{ \begin{array}{ll}
                = 0 & \mbox{return identity matrix}\\
                \ne 0 & \mbox{v must have a finite length}
                \end{array}
                \right.

        :seealso: :func:`~spatialmath.pose3d.SE3.angvec`, :func:`~spatialmath.pose3d.SE3.EulerVec`, :func:`~spatialmath.base.transforms3d.angvec2r`
        """
        return cls(base.angvec2tr(theta, v, unit=unit), check=False)

    @classmethod
    def AngVec(cls, theta, v, *, unit='rad'):
        r"""
        Create an SE(3) pure rotation matrix from rotation angle and axis

        :param θ: rotation
        :type θ: float
        :param unit: angular units: 'rad' [default], or 'deg'
        :type unit: str
        :param v: rotation axis, 3-vector
        :type v: array_like
        :return: SE(3) matrix
        :rtype: SE3 instance

        ``SE3.AngVec(θ, v)`` is an SE(3) rotation defined by
        a rotation of ``θ`` about the vector ``v``.

        .. deprecated:: 0.9.8
            Use :meth:`AngleAxis` instead.

        :seealso: :func:`~spatialmath.pose3d.SE3.angvec`, :func:`~spatialmath.pose3d.SE3.EulerVec`, :func:`~spatialmath.base.transforms3d.angvec2r`
        """
        return cls(base.angvec2tr(theta, v, unit=unit), check=False)

    @classmethod
    def EulerVec(cls, w):
        r"""
        Construct a new SE(3) pure rotation matrix from an Euler rotation vector

        :param ω: rotation axis
        :type ω: 3-element array_like
        :return: SE(3) rotation
        :rtype: SE3 instance

        ``SE3.EulerVec(ω)`` is a unit quaternion that describes the 3D rotation
        defined by a rotation of :math:`\theta = \lVert \omega \rVert` about the
        unit 3-vector :math:`\omega / \lVert \omega \rVert`.

        Example:

        .. runblock:: pycon
        
            >>> from spatialmath import SE3
            >>> SE3.EulerVec([0.5,0,0])

        .. note:: :math:`\theta = 0` the result in an identity matrix, otherwise
            ``V`` must have a finite length, ie. :math:`|V| > 0`.

        :seealso: :func:`~spatialmath.pose3d.SE3.AngVec`, :func:`~spatialmath.base.transforms3d.angvec2tr`
        """
        assert base.isvector(w, 3), 'w must be a 3-vector'
        w = base.getvector(w)
        theta = base.norm(w)
        return cls(base.angvec2tr(theta, w), check=False)

    @classmethod
    def Exp(cls, S, check=True):
        """
        Create an SE(3) matrix from se(3)

        :param S: Lie algebra se(3) matrix
        :type S: numpy ndarray
        :return: SE(3) matrix
        :rtype: SE3 instance

        - ``SE3.Exp(S)`` is an SE(3) rotation defined by its Lie algebra
          which is a 4x4 se(3) matrix (skew symmetric)
        - ``SE3.Exp(t)`` is an SE(3) rotation defined by a 6-element twist
          vector (the unique elements of the se(3) skew-symmetric matrix)

        :seealso: :func:`~spatialmath.base.transforms3d.trexp`, :func:`~spatialmath.base.transformsNd.skew`
        """
        if base.isvector(S, 6):
            return cls(base.trexp(base.getvector(S)), check=False)
        else:
            return cls(base.trexp(S), check=False)
            

    @classmethod
    def Delta(cls, d):
        r"""
        Create SE(3) from differential motion

        :param d: differential motion
        :type d: 6-element array_like
        :return: SE(3) matrix
        :rtype: SE3 instance


        ``SE3.Delta2tr(d)`` is an SE(3) representing differential 
        motion :math:`d = [\delta_x, \delta_y, \delta_z, \theta_x, \theta_y, \theta_z]`.

        :Reference: Robotics, Vision & Control: Second Edition, P. Corke, Springer 2016; p67.

        :seealso: :meth:`~delta` :func:`~spatialmath.base.transform3d.delta2tr`
        :SymPy: supported
        """
        return cls(base.trnorm(base.delta2tr(d)))

    @classmethod
    def Trans(cls, x, y=None, z=None):
        """
        Create SE(3) from translation vector

        :param x: x-coordinate or translation vector
        :type x: float or array_like(3)
        :param y: y-coordinate, defaults to None
        :type y: float, optional
        :param z: z-coordinate, defaults to None
        :type z: float, optional
        :return: SE(3) matrix
        :rtype: SE3 instance

        - ``SE3.Trans(x, y, z)`` is an SE(3) representing pure translation.

        - ``SE3.Trans([x, y, z])`` as above, but translation is given as an
          array.

        - ``SE3.Trans(t)`` where ``t`` is Nx3 then create an SE3 object with
          N elements whose translation is defined by the rows of ``t``.

        """
        if y is None and z is None:
            # single passed value, assume is 3-vector or Nx3
            t = base.getmatrix(x, (None, 3))
            return cls([base.transl(_t) for _t in t], check=False)
        else:
            return cls(np.array([x, y, z]))

    @classmethod
    def Tx(cls, x):
        """
        Create an SE(3) translation along the X-axis

        :param x: translation distance along the X-axis
        :type x: float
        :return: SE(3) matrix
        :rtype: SE3 instance

        `SE3.Tx(x)` is an SE(3) translation of ``x`` along the x-axis

        Example:

        .. runblock:: pycon

            >>> from spatialmath import SE3
            >>> SE3.Tx(2)
            >>> SE3.Tx([2,3])


        :seealso: :func:`~spatialmath.base.transforms3d.transl`
        :SymPy: supported
        """
        return cls([base.transl(_x, 0, 0) for _x in base.getvector(x)], check=False)


    @classmethod
    def Ty(cls, y):
        """
        Create an SE(3) translation along the Y-axis

        :param y: translation distance along the Y-axis
        :type y: float
        :return: SE(3) matrix
        :rtype: SE3 instance

        `SE3.Ty(y) is an SE(3) translation of ``y`` along the y-axis

        Example:

        .. runblock:: pycon

            >>> from spatialmath import SE3
            >>> SE3.Ty(2)
            >>> SE3.Ty([2,3])


        :seealso: :func:`~spatialmath.base.transforms3d.transl`
        :SymPy: supported
        """
        return cls([base.transl(0, _y, 0) for _y in base.getvector(y)], check=False)

    @classmethod
    def Tz(cls, z):
        """
        Create an SE(3) translation along the Z-axis

        :param z: translation distance along the Z-axis
        :type z: float
        :return: SE(3) matrix
        :rtype: SE3 instance

        `SE3.Tz(z)` is an SE(3) translation of ``z`` along the z-axis

        Example:

        .. runblock:: pycon

            >>> from spatialmath import SE3
            >>> SE3.Tz(2)
            >>> SE3.Tz([2,3])

        :seealso: :func:`~spatialmath.base.transforms3d.transl`
        :SymPy: supported
        """
        return cls([base.transl(0, 0, _z) for _z in base.getvector(z)], check=False)

    @classmethod
    def Rt(cls, R, t=None, check=True):
        """
        Create an SE(3) from rotation and translation

        :param R: rotation
        :type R: SO3 or ndarray(3,3)
        :param t: translation
        :type t: array_like(3)
        :param check: check rotation validity, defaults to True
        :type check: bool, optional
        :raises ValueError: bad rotation matrix
        :return: SE(3) matrix
        :rtype: SE3 instance
        """
        if isinstance(R, SO3):
            R = R.A
        elif base.isrot(R, check=check):
            pass
        else:
            raise ValueError('expecting SO3 or rotation matrix')

        if t is None:
            t = np.zeros((3,))
        return cls(base.rt2tr(R, t, check=check), check=check)

    def angdist(self, other, metric=6):
        r"""
        Angular distance metric between poses

        :param other: second rotation
        :type other: SE3 instance
        :param metric: metric, default is 6
        :type metric: int
        :raises TypeError: if other is not an SE3
        :return: angle in radians
        :rtype: float or ndarray

        ``T1.angdist(T2)`` is the geodesic norm, or geodesic distance between the
        rotational parts of the two poses.

        Several metrics are supported, the first 5 are computed after conversion
        to unit quaternions.

        ======   ===============================================================
        Metric   Details
        ======   ===============================================================
        0        :math:`1 - | \q_1 \bullet \q_2 | \in [0, 1]`
        1        :math:`\cos^{-1} | \q_1 \bullet \q_2 | \in [0, \pi/2]`
        2        :math:`\cos^{-1} | \q_1 \bullet \q_2 | \in [0, \pi/2]`
        3        :math:`2 \tan^{-1} \| \q_1 - \q_2\| / \|\q_1 + \q_2\| \in [0, \pi/2]`
        4        :math:`\cos^{-1} \left( 2 (\q_1 \bullet \q_2)^2 - 1\right) \in [0, 1]`
        5        :math:`\|I - \mat{R}_1 \mat{R}_2^T\| \in [0, 2]`
        6        :math:`\|\log \mat{R}_1 \mat{R}_2^T\| \in [0, \pi]`
        ======   ===============================================================

        Example:

        .. runblock:: pycon

            >>> from spatialmath import SE3
            >>> T1 = SE3.Rx(0.3)
            >>> T2 = SE3.Ry(0.3)
            >>> print(T1.angdist(T1))
            >>> print(T1.angdist(T2))

        .. note::
            - metrics 1, 2, 4 can throw ValueError "math domain error" due to
              numeric errors which push the argument of ``acos()`` marginally
              outside its domain [0, 1].
            - metrics 2 and 3 are equivalent, but 3 is more robust

        :seealso: :func:`UnitQuaternion.angdist`
        """

        if metric < 5:
            from spatialmath.quaternion import UnitQuaternion

            return UnitQuaternion(self).angdist(UnitQuaternion(other), metric=metric)

        elif metric == 5:
            op = lambda T1, T2: np.linalg.norm(np.eye(3) - T1[:3,:3] @ T2[:3,:3].T)
        elif metric == 6:
            op = lambda T1, T2: base.norm(base.trlog(T1[:3,:3] @ T2[:3,:3].T, twist=True))
        else:
            raise ValueError('unknown metric')
        
        ad = self._op2(other, op)
        if isinstance(ad, list):
            return np.array(ad)
        else:
            return ad

    # @classmethod
    # def SO3(cls, R, t=None, check=True):
    #     if isinstance(R, SO3):
    #         R = R.A
    #     elif base.isrot(R, check=check):
    #         pass
    #     else:
    #         raise ValueError('expecting SO3 or rotation matrix')
    #     if t is None:
    #         return cls(base.r2t(R))
    #     else:
    #         return cls(base.rt2tr(R, t))

if __name__ == '__main__':   # pragma: no cover

    import pathlib
    exec(open(pathlib.Path(__file__).parent.parent.absolute() / "tests" / "test_pose3d.py").read())  # pylint: disable=exec-used



import numpy.testing as nt
import matplotlib.pyplot as plt
import unittest

"""
we will assume that the primitives rotx,trotx, etc. all work
"""
from math import pi
from spatialmath.pose2d import *
# from spatialmath import super_pose as sp
from spatialmath.base import *
import spatialmath.base.argcheck as argcheck
import spatialmath as sm
from spatialmath.baseposematrix import BasePoseMatrix
from spatialmath.twist import BaseTwist

def array_compare(x, y):
    if isinstance(x, BasePoseMatrix):
        x = x.A
    if isinstance(y, BasePoseMatrix):
        y = y.A
    if isinstance(x, BaseTwist):
        x = x.S
    if isinstance(y, BaseTwist):
        y = y.S
    nt.assert_array_almost_equal(x, y)


class TestSO2(unittest.TestCase):

    @classmethod
    def tearDownClass(cls):
        plt.close('all')

    def test_constructor(self):
        
        
        # null case
        x = SO2()
        self.assertIsInstance(x, SO2)
        self.assertEqual(len(x), 1)
        array_compare(x.A, np.eye(2,2))
        
        ## from angle
        
        array_compare(SO2(0).A, np.eye(2))
        array_compare(SO2(pi / 2).A, rot2(pi / 2))
        array_compare(SO2(90, unit='deg').A, rot2(pi / 2))
       
        ## from R
        
        array_compare(SO2(np.eye(2,2)).A, np.eye(2,2))
    
        array_compare(SO2( rot2(pi / 2)).A, rot2(pi / 2))
        array_compare(SO2( rot2(pi)).A, rot2(pi))
           
        
        ## R,T
        array_compare(SO2( np.eye(2)).R, np.eye(2))
       
        array_compare(SO2( rot2(pi / 2)).R, rot2(pi / 2))
        
        
        ## vectorised forms of R
        R = SO2.Empty()
        for theta in [-pi / 2, 0, pi / 2, pi]:
            R.append(SO2(theta))
        self.assertEqual(len(R), 4)
        array_compare(R[0], rot2(-pi / 2))
        array_compare(R[3], rot2(pi))

        # TODO self.assertEqual(SO2(R).R, R)
        
        ## copy constructor
        r = SO2(0.3)
        c = SO2(r)
        array_compare(r, c)
        r = SO2(0.4)
        array_compare(c, SO2(0.3))
    
    def test_concat(self):
        x = SO2()
        xx = SO2([x, x, x, x])
        
        self.assertIsInstance(xx, SO2)
        self.assertEqual(len(xx), 4)
    
    def test_primitive_convert(self):
        # char
        
        s = str( SO2())
        self.assertIsInstance(s, str)
    
    def test_shape(self):
        a = SO2()
        self.assertEqual(a._A.shape, a.shape)

    def test_constructor_Exp(self):

        array_compare(SO2.Exp( skew(0.3)).R, rot2(0.3))
        array_compare(SO2.Exp( 0.3).R, rot2(0.3))

        x = SO2.Exp([0, 0.3, 1])
        self.assertEqual(len(x), 3)
        array_compare(x[0], rot2(0))
        array_compare(x[1], rot2(0.3))
        array_compare(x[2], rot2(1))

        x = SO2.Exp([skew(x) for x in [0, 0.3, 1]])
        self.assertEqual(len(x), 3)
        array_compare(x[0], rot2(0))
        array_compare(x[1], rot2(0.3))
        array_compare(x[2], rot2(1))
    
    def test_isa(self):
        
        self.assertTrue(SO2.isvalid(rot2(0)))
    
        self.assertFalse(SO2.isvalid(1))
    
    def test_resulttype(self):
        
        r = SO2()
        self.assertIsInstance(r, SO2)
    
        self.assertIsInstance(r * r, SO2)
        
    
        self.assertIsInstance(r / r, SO2)
        
        self.assertIsInstance(r.inv(), SO2)
    
    
    def test_multiply(self):
        
        vx = np.r_[1, 0]
        vy = np.r_[0, 1]
        
        r0 = SO2(0)
        r1 = SO2(pi / 2)
        r2 = SO2(pi)
        u = SO2()
        
        ## SO2-SO2, product
        # scalar x scalar
        
        array_compare(r0 * u, r0)
        array_compare(u * r0, r0)
        
        # vector x vector
        array_compare(SO2([r0, r1, r2]) * SO2([r2, r0, r1]), SO2([r0 * r2, r1 * r0, r2 * r1]))
        
        # scalar x vector
        array_compare(r1 * SO2([r0, r1, r2]), SO2([r1 * r0, r1 * r1, r1 * r2]))
        
        # vector x scalar
        array_compare(SO2([r0, r1, r2]) * r2, SO2([r0 * r2, r1 * r2, r2 * r2]))
        
        ## SO2-vector product
        # scalar x scalar
        
        array_compare(r1 * vx, np.c_[vy])
        
        # vector x vector
        #array_compare(SO2([r0, r1, r0]) * np.c_[vy, vx, vx], np.c_[vy, vy, vx])
        
        # scalar x vector
        array_compare(r1 * np.c_[vx, vy, -vx], np.c_[vy, -vx, -vy])
        
        # vector x scalar
        array_compare(SO2([r0, r1, r2]) * vy, np.c_[vy, -vx, -vy])

    
    def test_divide(self):
        
        r0 = SO2(0)
        r1 = SO2(pi / 2)
        r2 = SO2(pi)
        u = SO2()
        
        # scalar / scalar
        # implicity tests inv
    
        array_compare(r1 / u, r1)
        array_compare(r1 / r1, u)
    
        # vector / vector
        array_compare(SO2([r0, r1, r2]) / SO2([r2, r1, r0]), SO2([r0 / r2, r1 / r1, r2 / r0]))
        
        # vector / scalar
        array_compare(SO2([r0, r1, r2]) / r1, SO2([r0 / r1, r1 / r1, r2 / r1]))
    
    
    def test_conversions(self):
        
        T = SO2(pi / 2).SE2()

        self.assertIsInstance(T, SE2)
    
        
        ## Lie stuff
        th = 0.3
        RR = SO2(th)
        array_compare(RR.log(), skew(th))
    
    
    def test_miscellany(self):
        
        r = SO2( 0.3,)
        self.assertAlmostEqual(np.linalg.det(r.A), 1)
        
        self.assertEqual(r.N, 2)
        
        self.assertFalse(r.isSE)
    
    
    def test_printline(self):
        
        R = SO2( 0.3)
        
        R.printline()
        # s = R.printline(file=None)
        # self.assertIsInstance(s, str)

        R = SO2([0.3, 0.4, 0.5])
        s = R.printline(file=None)
        # self.assertIsInstance(s, str)
        # self.assertEqual(s.count('\n'), 2)
        
    def test_plot(self):
        plt.close('all')
        
        R = SO2( 0.3)
        R.plot(block=False)
        
        R2 = SO2(0.6)
        # R.animate()
        # R.animate(start=R2)
        
        
# ============================== SE2 =====================================#

class TestSE2(unittest.TestCase):

    @classmethod
    def tearDownClass(cls):
        plt.close('all')

    def test_constructor(self):
        
        self.assertIsInstance(SE2(), SE2)
    
        ## null
        array_compare(SE2().A, np.eye(3,3))
        
        # from x,y
        x = SE2(2, 3)
        self.assertIsInstance(x, SE2)
        self.assertEqual(len(x), 1)
        array_compare(x.A, np.array([[1,0,2],[0,1,3],[0,0,1]]))
        
        x = SE2([2, 3])
        self.assertIsInstance(x, SE2)
        self.assertEqual(len(x), 1)
        array_compare(x.A, np.array([[1,0,2],[0,1,3],[0,0,1]]))

        # from x,y,theta
        x = SE2(2, 3, pi / 2)
        self.assertIsInstance(x, SE2)
        self.assertEqual(len(x), 1)
        array_compare(x.A, np.array([[0,-1,2],[1,0,3],[0,0,1]]))
        
        x = SE2([2, 3, pi / 2])
        self.assertIsInstance(x, SE2)
        self.assertEqual(len(x), 1)
        array_compare(x.A, np.array([[0,-1,2],[1,0,3],[0,0,1]]))
        
        x = SE2(2, 3, 90, unit='deg')
        self.assertIsInstance(x, SE2)
        self.assertEqual(len(x), 1)
        array_compare(x.A, np.array([[0,-1,2],[1,0,3],[0,0,1]]))
        
        x = SE2([2, 3, 90], unit='deg')
        self.assertIsInstance(x, SE2)
        self.assertEqual(len(x), 1)
        array_compare(x.A, np.array([[0,-1,2],[1,0,3],[0,0,1]]))
    
        
        ## T
        T = transl2(1, 2) @ trot2(0.3)
        x = SE2(T)
        self.assertIsInstance(x, SE2)
        self.assertEqual(len(x), 1)
        array_compare(x.A, T)
        
        
        ## copy constructor
        TT = SE2(x)
        array_compare(SE2(TT).A, T)
        x = SE2()
        array_compare(SE2(TT).A, T)
        
        ## vectorised versions
        
        T1 = transl2(1,2) @ trot2(0.3)
        T2 = transl2(1,-2) @ trot2(-0.4)
        
        x =SE2([T1, T2, T1, T2])
        self.assertIsInstance(x, SE2)
        self.assertEqual(len(x), 4)
        array_compare(x[0], T1)
        array_compare(x[1], T2)

    def test_shape(self):
        a = SE2()
        self.assertEqual(a._A.shape, a.shape)

    def test_concat(self):
        x = SE2()
        xx = SE2([x, x, x, x])
        
        self.assertIsInstance(xx, SE2)
        self.assertEqual(len(xx), 4)
    
    
    def test_constructor_Exp(self):

        array_compare(SE2.Exp(skewa([1,2,0])), transl2(1,2))
        array_compare(SE2.Exp(np.r_[1,2,0]), transl2(1,2))

        x = SE2.Exp([(1,2,0), (3,4,0), (5,6,0)])
        self.assertEqual(len(x), 3)
        array_compare(x[0], transl2(1,2))
        array_compare(x[1], transl2(3,4))
        array_compare(x[2], transl2(5,6))

        x = SE2.Exp([skewa(x) for x in [(1,2,0), (3,4,0), (5,6,0)]])
        self.assertEqual(len(x), 3)
        array_compare(x[0], transl2(1,2))
        array_compare(x[1], transl2(3,4))
        array_compare(x[2], transl2(5,6))
    
    def test_isa(self):
        
        self.assertTrue(SE2.isvalid(trot2(0)))
        self.assertFalse(SE2.isvalid(1))

    
    def test_resulttype(self):
        
        t = SE2()
        self.assertIsInstance(t, SE2)
        self.assertIsInstance(t * t, SE2)
        self.assertIsInstance(t / t, SE2)
        self.assertIsInstance(t.inv(), SE2)
        self.assertIsInstance(t + t, np.ndarray)
        self.assertIsInstance(t + 1, np.ndarray)
        self.assertIsInstance(t - 1, np.ndarray)
        self.assertIsInstance(1 + t, np.ndarray)
        self.assertIsInstance(1 - t, np.ndarray)
        self.assertIsInstance(2 * t, np.ndarray)
        self.assertIsInstance(t * 2, np.ndarray)

    
    def test_inverse(self):    
        
        T1 = transl2(1, 2) @ trot2(0.3)
        TT1 = SE2(T1)
        
        # test inverse
        array_compare(TT1.inv().A, np.linalg.inv(T1))
        
        array_compare(TT1 * TT1.inv(),  np.eye(3))
        array_compare(TT1.inv() * TT1, np.eye(3))
        
        # vector case
        TT2 = SE2([TT1, TT1])
        u = [np.eye(3), np.eye(3)]
        array_compare(TT2.inv() * TT1, u)
    
    
    def test_Rt(self):
       
        
        TT1 = SE2.Rand()
        T1 = TT1.A
        R1 = t2r(T1)
        t1 = transl2(T1)
        
        array_compare(TT1.A, T1)
        array_compare(TT1.R, R1)
        array_compare(TT1.t, t1)
        
        TT = SE2([TT1, TT1, TT1])
        array_compare(TT.t, [t1, t1, t1])
    
    
    def test_arith(self):
        
        
        TT1 = SE2.Rand()
        T1 = TT1.A
        TT2 = SE2.Rand()
        T2 = TT2.A
    
        I = SE2()
        
        ## SE2, * SE2, product
        # scalar x scalar
        
        array_compare(TT1 * TT2, T1 @ T2)
        array_compare(TT2 * TT1, T2 @ T1)
        array_compare(TT1 * I, T1)
        array_compare(TT2 * I, TT2)

        
        # vector x vector
        array_compare(SE2([TT1, TT1, TT2]) * SE2([TT2, TT1, TT1]), SE2([TT1*TT2, TT1*TT1, TT2*TT1]))
        
        # scalar x vector
        array_compare(TT1 * SE2([TT2, TT1]), SE2([TT1*TT2, TT1*TT1]))
        
        # vector x scalar
        array_compare(SE2([TT1, TT2]) * TT2, SE2([TT1*TT2, TT2*TT2]))
        
        ## SE2, * vector product
        vx = np.r_[1, 0]
        vy = np.r_[0, 1]
    
        # scalar x scalar
        
        array_compare(TT1 * vy, h2e( T1 @ e2h(vy)))
        
        # # vector x vector
        # array_compare(SE2([TT1, TT2]) * np.c_[vx, vy], np.c_[h2e(T1 @ e2h(vx)), h2e(T2 @ e2h(vy))])
        
        # scalar x vector
        array_compare(TT1 * np.c_[vx, vy], h2e( T1 @ e2h(np.c_[vx, vy])))
        
        # vector x scalar
        array_compare(SE2([TT1, TT2, TT1]) * vy, np.c_[h2e(T1 @ e2h(vy)), h2e(T2 @ e2h(vy)), h2e(T1 @ e2h(vy))])
    
    def test_defs(self):
        
        # log
        # x = SE2.Exp([2, 3, 0.5])
        # array_compare(x.log(), np.array([[0, -0.5, 2], [0.5, 0, 3], [0, 0, 0]]))
        pass
    
    def test_conversions(self):
        
        
        ##  SE2,                     convert to SE2, class
    
        TT = SE2(1, 2, 0.3)
        
        array_compare(TT, transl2(1, 2) @ trot2(0.3))
        
        ## xyt
        array_compare(TT.xyt(), np.r_[1, 2, 0.3])
        
        ## Lie stuff
        x = TT.log()
        self.assertTrue(isskewa(x))

    
    def test_interp(self):
        TT = SE2(2, -4, 0.6)
        I = SE2()
        
        z = I.interp(TT, s=0)
        self.assertIsInstance(z, SE2)
        
        array_compare(I.interp(TT, s=0), I)
        array_compare(I.interp(TT, s=1), TT)
        array_compare(I.interp(TT, s=0.5), SE2(1, -2, 0.3))
    
    def test_miscellany(self):
        
        TT = SE2(1, 2, 0.3)
        
        self.assertEqual(TT.A.shape, (3,3))
            
        self.assertTrue(TT.isSE)
        
        self.assertIsInstance(TT, SE2)
    
    def test_display(self):
        
        T1 = SE2.Rand()
        
        T1.printline()
        
    def test_graphics(self):
        
        plt.close('all')
        T1 = SE2.Rand()
        T2 = SE2.Rand()
        
        T1.plot(block=False, dims=[-2,2])
        
        T1.animate(repeat=False, dims=[-2,2], nframes=10)
        T1.animate(T0=T2, repeat=False, dims=[-2,2], nframes=10)


# ---------------------------------------------------------------------------------------#
if __name__ == '__main__':

    unittest.main(buffer=True)



import math
from math import pi
import numpy as np

import numpy.testing as nt
import unittest

from spatialmath import DualQuaternion, UnitDualQuaternion, Quaternion, SE3
from spatialmath import base


def qcompare(x, y):
    if isinstance(x, Quaternion):
        x = x.vec
    elif isinstance(x, SMPose):
        x = x.A
    if isinstance(y, Quaternion):
        y = y.vec
    elif isinstance(y, SMPose):
        y = y.A
    nt.assert_array_almost_equal(x, y)

class TestDualQuaternion(unittest.TestCase):

    def test_init(self):

        dq = DualQuaternion(Quaternion([1.,2,3,4]), Quaternion([5.,6,7,8]))
        nt.assert_array_almost_equal(dq.vec, np.r_[1,2,3,4,5,6,7,8])

        dq = DualQuaternion([1.,2,3,4,5,6,7,8])
        nt.assert_array_almost_equal(dq.vec, np.r_[1,2,3,4,5,6,7,8])
        dq = DualQuaternion(np.r_[1,2,3,4,5,6,7,8])
        nt.assert_array_almost_equal(dq.vec, np.r_[1,2,3,4,5,6,7,8])

    def test_pure(self):

        dq = DualQuaternion.Pure([1.,2,3])
        nt.assert_array_almost_equal(dq.vec, np.r_[1,0,0,0,  0,1,2,3])

    def test_strings(self):

        dq = DualQuaternion(Quaternion([1.,2,3,4]), Quaternion([5.,6,7,8]))
        self.assertIsInstance(str(dq), str)
        self.assertIsInstance(repr(dq), str)

    def test_conj(self):
        dq = DualQuaternion(Quaternion([1.,2,3,4]), Quaternion([5.,6,7,8]))
        nt.assert_array_almost_equal(dq.conj().vec, np.r_[1,-2,-3,-4, 5,-6,-7,-8])

    # def test_norm(self):
    #     q1 = Quaternion([1.,2,3,4])
    #     q2 = Quaternion([5.,6,7,8])

    #     dq = DualQuaternion(q1, q2)
    #     nt.assert_array_almost_equal(dq.norm(), (q1.norm(), q2.norm()))

    def test_plus(self):
        dq = DualQuaternion(Quaternion([1.,2,3,4]), Quaternion([5.,6,7,8]))
        s = dq + dq
        nt.assert_array_almost_equal(s.vec, 2*np.r_[1,2,3,4,5,6,7,8])

    def test_minus(self):
        dq = DualQuaternion(Quaternion([1.,2,3,4]), Quaternion([5.,6,7,8]))
        s = dq - dq
        nt.assert_array_almost_equal(s.vec, np.zeros((8,)))

    def test_matrix(self):

        dq1 = DualQuaternion(Quaternion([1.,2,3,4]), Quaternion([5.,6,7,8]))

        M = dq1.matrix()
        self.assertIsInstance(M, np.ndarray)
        self.assertEqual(M.shape, (8,8))

    def test_multiply(self):
        dq1 = DualQuaternion(Quaternion([1.,2,3,4]), Quaternion([5.,6,7,8]))
        dq2 = DualQuaternion(Quaternion([4,3,2,1]), Quaternion([5,6,7,8]))

        M = dq1.matrix()
        v = dq2.vec
        nt.assert_array_almost_equal(M @ v, (dq1 * dq2).vec)

    def test_unit(self):
        pass


class TestUnitDualQuaternion(unittest.TestCase):

    def test_init(self):

        T = SE3.Rx(pi/4)
        dq = UnitDualQuaternion(T)
        nt.assert_array_almost_equal(dq.SE3().A, T.A)

    def test_norm(self):
        T = SE3.Rx(pi/4)
        dq = UnitDualQuaternion(T)
        nt.assert_array_almost_equal(dq.norm(), (1,0))

    def test_multiply(self):
        T1 = SE3.Rx(pi/4)
        T2 = SE3.Rz(-pi/3)

        T = T1 * T2

        d1 = UnitDualQuaternion(T1)
        d2 = UnitDualQuaternion(T2)

        d = d1 * d2
        nt.assert_array_almost_equal(d.SE3().A, T.A)


# ---------------------------------------------------------------------------------------#
if __name__ == '__main__':  # pragma: no cover

    unittest.main()



# Part of Spatial Math Toolbox for Python
# Copyright (c) 2000 Peter Corke
# MIT Licence, see details in top-level file: LICENCE

"""
This package provides a light-weight wrapper to support use of SymPy.  It
generalizes some common functions so that they can accept numerical or
Symbolic arguments.

If SymPy is not installed then only the standard numeric operations are
supported. 
"""

import math

try:  # pragma: no cover
    # print('Using SymPy')
    import sympy

    _symbolics = True
    symtype = (sympy.Expr,)

except ImportError:  # pragma: no cover
    _symbolics = False
    symtype = ()


# ---------------------------------------------------------------------------------------#


def symbol(name, real=True):
    """
    Create symbolic variables

    :param name: symbol names
    :type name: str
    :param real: assume variable is real, defaults to True
    :type real: bool, optional
    :return: SymPy symbols
    :rtype: sympy

    .. runblock:: pycon

        >>> from spatialmath.base.symbolic import *
        >>> theta = symbol('theta')
        >>> theta
        >>> theta, psi = symbol('theta psi')
        >>> theta
        >>> psi
        >>> q = symbol('q_:6')
        >>> q

    .. note:: In Jupyter symbols are pretty printed.

        - symbols named after greek letters will appear as greek letters
        - underscore means subscript as it does in LaTex, so the symbols ``q``
          above will be subscripted.

    :seealso: :func:`sympy.symbols`
    """
    return sympy.symbols(name, real=real)


def issymbol(var):
    """
    Test if variable is symbolic

    :param var: variable to test
    :return: whether variable is symbolic
    :rtype: bool

    .. runblock:: pycon

        >>> from spatialmath.base.symbolic import *
        >>> theta = symbol('theta')
        >>> issymbol(theta)
        >>> issymbol(3.4)

    """
    if _symbolics:
        if isinstance(var, (list, tuple)):
            return any([isinstance(x, symtype) for x in var])
        else:
            return isinstance(var, symtype)
    else:
        return False


def sin(theta):
    """
    Generalized sine function

    :param θ: argument
    :type θ: float or symbolic
    :return: sin(θ)
    :rtype: float or symbolic

    .. runblock:: pycon

        >>> from spatialmath.base.symbolic import *
        >>> theta = symbol('theta')
        >>> sin(theta)
        >>> sin(0.5)

    :seealso: :func:`sympy.sin`
    """
    if issymbol(theta):
        return sympy.sin(theta)
    else:
        return math.sin(theta)


def cos(theta):
    """
    Generalized cosine function

    :param θ: argument
    :type θ: float or symbolic
    :return: cos(θ)
    :rtype: float or symbolic

    .. runblock:: pycon

        >>> from spatialmath.base.symbolic import *
        >>> theta = symbol('theta')
        >>> cos(theta)
        >>> cos(0.5)

    :seealso: :func:`sympy.cos`
    """
    if issymbol(theta):
        return sympy.cos(theta)
    else:
        return math.cos(theta)

def tan(theta):
    """
    Generalized tangent function

    :param θ: argument
    :type θ: float or symbolic
    :return: tan(θ)
    :rtype: float or symbolic

    .. runblock:: pycon

        >>> from spatialmath.base.symbolic import *
        >>> theta = symbol('theta')
        >>> tan(theta)
        >>> tan(0.5)

    :seealso: :func:`sympy.cos`
    """
    if issymbol(theta):
        return sympy.tan(theta)
    else:
        return math.tan(theta)

def sqrt(v):
    """
    Generalized sqrt function

    :param v: argument
    :type v: float or symbolic
    :return: √ v
    :rtype: float or symbolic

    .. runblock:: pycon

        >>> from spatialmath.base.symbolic import *
        >>> x = symbol('x')
        >>> sqrt(x ** 2)
        >>> sqrt(4)

    :seealso: :func:`sympy.sqrt`
    """
    if issymbol(v):
        return sympy.sqrt(v)
    else:
        return math.sqrt(v)


def zero():
    """
    Symbolic constant: zero

    :return: 0
    :rtype: symbolic

    .. runblock:: pycon

        >>> from spatialmath.base.symbolic import *
        >>> x = symbol('x')
        >>> zero()
        >>> x + zero()

    :seealso: :func:`sympy.S.Zero`
    """
    return sympy.S.Zero


def one():
    """
    Symbolic constant: one

    :return: 1
    :rtype: symbolic

    .. runblock:: pycon

        >>> from spatialmath.base.symbolic import *
        >>> x = symbol('x')
        >>> one()
        >>> one() * x

    :seealso: :func:`sympy.S.One`
    """
    return sympy.S.One


def negative_one():
    """
    Symbolic constant: negative one

    :return: -1
    :rtype: symbolic

    .. runblock:: pycon

        >>> from spatialmath.base.symbolic import *
        >>> x = symbol('x')
        >>> negative_one()
        >>> negative_one() * x

    :seealso: :func:`sympy.S.NegativeOne`
    """
    return sympy.S.NegativeOne


def pi():
    """
    Symbolic constant: pi

    :return: π
    :rtype: symbolic

    .. runblock:: pycon

        >>> from spatialmath.base.symbolic import *
        >>> import math
        >>> sin(pi())
        >>> sin(math.pi)

    :seealso: :func:`sympy.S.Pi`
    """
    return sympy.S.Pi


def simplify(x):
    """
    Symbolic simplification

    :param x: expression to simplify
    :type x: symbolic
    :return: -1
    :rtype: symbolic

    .. runblock:: pycon

        >>> from spatialmath.base.symbolic import *
        >>> x = symbol('x')
        >>> y = (x - 1) * (x + 1) - x ** 2
        >>> y
        >>> simplify(y)

    :seealso: :func:`sympy.simplify`
    """
    if _symbolics:
        return sympy.simplify(x)
    else:
        return x


def det(x):
    """
    Symbolic determinant

    :param m: matrix
    :type x: ndarray with symbolic elements
    :return: determinant
    :rtype: ndarray with symbolic elements

    .. runblock:: pycon

        >>> from spatialmath.base.symbolic import *
        >>> from spatialmath.base import rot2
        >>> theta = symbol('theta')
        >>> R = rot2(theta)
        >>> print(R)
        >>> print(det(R))
        >>> simplify(print(det(R)))

    .. note:: Converts to a SymPy ``Matrix`` and then back again.
    """

    return sympy.Matrix(x).det()



import math
from itertools import product
import warnings
import numpy as np
import scipy as sp
from matplotlib import colors

from spatialmath import base as smbase

try:
    import matplotlib.pyplot as plt
    from matplotlib.patches import Circle
    from mpl_toolkits.mplot3d.art3d import (
        Poly3DCollection,
        Line3DCollection,
        pathpatch_2d_to_3d,
    )

    _matplotlib_exists = True
except ImportError:  # pragma: no cover
    _matplotlib_exists = False

"""
Set of functions to draw 2D and 3D graphical primitives using matplotlib.

The 2D functions all allow color and line style to be specified by a fmt string
like, 'r' or 'b--'.

The 3D functions require explicity arguments to set properties, like color='b'

All return a list of the graphic objects they create.

"""
# TODO
# return a redrawer object, that can be used for animation

# =========================== 2D shapes =================================== #


def plot_text(pos, text=None, ax=None, color=None, **kwargs):
    """
    Plot text using matplotlib

    :param pos: position of text
    :type pos: array_like(2)
    :param text: text
    :type text: str
    :param ax: axes to draw in, defaults to ``gca()``
    :type ax: Axis, optional
    :param color: text color, defaults to None
    :type color: str or array_like(3), optional
    :param kwargs: additional arguments passed to ``pyplot.text()``
    :return: the matplotlib object
    :rtype: list of Text instance

    Example:

    .. runblock:: pycon

        >>> from spatialmath.base import plotvol2, plot_text
        >>> plotvol2(5)
        >>> plot_text((1,3), 'foo')
        >>> plot_text((2,2), 'bar', 'b')
        >>> plot_text((2,2), 'baz', fontsize=14, horizontalalignment='centre')
    """

    defaults = {"horizontalalignment": "left", "verticalalignment": "center"}
    for k, v in defaults.items():
        if k not in kwargs:
            kwargs[k] = v
    if ax is None:
        ax = plt.gca()

    handle = ax.text(pos[0], pos[1], text, color=color, **kwargs)
    return [handle]


def plot_point(pos, marker="bs", text=None, ax=None, textargs=None, textcolor=None, **kwargs):
    """
    Plot a point using matplotlib

    :param pos: position of marker
    :type pos: array_like(2), ndarray(2,n), list of 2-tuples
    :param marker: matplotlub marker style, defaults to 'bs'
    :type marker: str or list of str, optional
    :param text: text label, defaults to None
    :type text: str, optional
    :param ax: axes to plot in, defaults to ``gca()``
    :type ax: Axis, optional
    :return: the matplotlib object
    :rtype: list of Text and Line2D instances

    Plot one or more points, with optional text label.

    - The color of the marker can be different to the color of the text, the
      marker color is specified by a single letter in the marker string.

    - A point can have multiple markers, given as a list, which will be
      overlaid, for instance ``["rx", "ro"]`` will give a ⨂ symbol.

    - The optional text label is placed to the right of the marker, and
      vertically aligned.

    - Multiple points can be marked if ``pos`` is a 2xn array or a list of
      coordinate pairs.  In this case:
        - all points have the same ``text`` label
        - ``text`` can include the format string {} which is susbstituted for the
          point index, starting at zero
        - ``text`` can be a tuple containing a format string followed by vectors
          of shape(n).  For example::

            ``("#{0} a={1:.1f}, b={2:.1f}", a, b)``

          will label each point with its index (argument 0) and consecutive
          elements of ``a`` and ``b`` which are arguments 1 and 2 respectively.

    Examples:

    - ``plot_point((1,2))`` plot default marker at coordinate (1,2)
    - ``plot_point((1,2), 'r*')`` plot red star at coordinate (1,2)
    - ``plot_point((1,2), 'r*', 'foo')`` plot red star at coordinate (1,2) and
      label it as 'foo'
    - ``plot_point(p, 'r*')`` plot red star at points defined by columns of
      ``p``.
    - ``plot_point(p, 'r*', 'foo')`` plot red star at points defined by columns
      of ``p`` and label them all as 'foo'
    - ``plot_point(p, 'r*', '{0}')`` plot red star at points defined by columns
      of ``p`` and label them sequentially from 0
    - ``plot_point(p, 'r*', ('{1:.1f}', z))`` plot red star at points defined by
      columns of ``p`` and label them all with successive elements of ``z``.
    """

    if isinstance(pos, np.ndarray):
        if pos.ndim == 1:
            x = pos[0]
            y = pos[1]
        elif pos.ndim == 2 and pos.shape[0] == 2:
            x = pos[0, :]
            y = pos[1, :]
    elif isinstance(pos, (tuple, list)):
        # [x, y]
        # [(x,y), (x,y), ...]
        # [xlist, ylist]
        # [xarray, yarray]
        if smbase.islistof(pos, (tuple, list)):
            x = [z[0] for z in pos]
            y = [z[1] for z in pos]
        elif smbase.islistof(pos, np.ndarray):
            x = pos[0]
            y = pos[1]
        else:
            x = pos[0]
            y = pos[1]

    textopts = {
        "fontsize": 12,
        "horizontalalignment": "left",
        "verticalalignment": "center",
    }
    if textargs is not None:
        textopts = {**textopts, **textargs}
    if textcolor is not None and "color" not in textopts:
        textopts["color"] = textcolor

    if ax is None:
        ax = plt.gca()

    handles = []
    if isinstance(marker, (list, tuple)):
        for m in marker:
            handles.append(ax.plot(x, y, m, **kwargs))
    else:
        handles.append(ax.plot(x, y, marker, **kwargs))
    if text is not None:
        try:
            xy = zip(x, y)
        except TypeError:
            xy = [(x, y)]
        if isinstance(text, str):
            # simple string, but might have format chars
            for i, (x, y) in enumerate(xy):
                handles.append(ax.text(x, y, " " + text.format(i), **textopts))
        elif isinstance(text, (tuple, list)):
            for i, (x, y) in enumerate(xy):
                handles.append(
                    ax.text(
                        x,
                        y,
                        " " + text[0].format(i, *[d[i] for d in text[1:]]),
                        **textopts
                    )
                )
    return handles


def plot_homline(lines, *args, ax=None, xlim=None, ylim=None, **kwargs):
    r"""
    Plot a homogeneous line using matplotlib

    :param lines: homgeneous lines
    :type lines: array_like(3), ndarray(3,N)
    :param ax: axes to plot in, defaults to ``gca()``
    :type ax: Axis, optional
    :param kwargs: arguments passed to ``plot``
    :return: matplotlib object
    :rtype: list of Line2D instances

    Draws the 2D line given in homogeneous form :math:`\ell[0] x + \ell[1] y + \ell[2] = 0` in the current
    2D axes.

    .. warning: A set of 2D axes must exist in order that the axis limits can
        be obtained. The line is drawn from edge to edge.

    If ``lines`` is a 3xN array then ``N`` lines are drawn, one per column.

    Example:

    .. runblock:: pycon

        >>> from spatialmath.base import plotvol2, plot_homline
        >>> plotvol2(5)
        >>> plot_homline((1, -2, 3))
        >>> plot_homline((1, -2, 3), 'k--') # dashed black line
    """
    ax = axes_logic(ax, 2)
    # get plot limits from current graph
    if xlim is None:
        xlim = np.r_[ax.get_xlim()]
    if ylim is None:
        ylim = np.r_[ax.get_ylim()]

    # if lines.ndim == 1:
    #     lines = lines.
    lines = smbase.getmatrix(lines, (3, None))

    handles = []
    for line in lines.T:  # for each column
        if abs(line[1]) > abs(line[0]):
            y = (-line[2] - line[0] * xlim) / line[1]
            ax.plot(xlim, y, *args, **kwargs)
        else:
            x = (-line[2] - line[1] * ylim) / line[0]
            handles.append(ax.plot(x, ylim, *args, **kwargs))

    return handles


def plot_box(
    *fmt,
    lbrt=None,
    lrbt=None,
    lbwh=None,
    bbox=None,
    ltrb=None,
    lb=None,
    lt=None,
    rb=None,
    rt=None,
    wh=None,
    centre=None,
    w=None,
    h=None,
    ax=None,
    filled=False,
    **kwargs
):
    """
    Plot a 2D box using matplotlib

    :param bl: bottom-left corner, defaults to None
    :type bl: array_like(2), optional
    :param tl: top-left corner, defaults to None
    :type tl: [array_like(2), optional
    :param br: bottom-right corner, defaults to None
    :type br: array_like(2), optional
    :param tr: top-right corner, defaults to None
    :type tr: array_like(2), optional
    :param wh: width and height, if both are the same provide scalar, defaults to None
    :type wh: scalar, array_like(2), optional
    :param centre: centre of box, defaults to None
    :type centre: array_like(2), optional
    :param w: width of box, defaults to None
    :type w: float, optional
    :param h: height of box, defaults to None
    :type h: float, optional
    :param ax: the axes to draw on, defaults to ``gca()``
    :type ax: Axis, optional
    :param bbox: bounding box matrix, defaults to None
    :type bbox: ndarray(2,2), optional
    :param color: box outline color
    :type color: array_like(3) or str
    :param fillcolor: box fill color
    :type fillcolor: array_like(3) or str
    :param alpha: transparency, defaults to 1
    :type alpha: float, optional
    :param thickness: line thickness, defaults to None
    :type thickness: float, optional
    :return: the matplotlib object
    :rtype: list of Line2D or Patch.Rectangle instance

    The box can be specified in many ways:

    - bounding box which is a 2x2 matrix [xmin, xmax, ymin, ymax]
    - bounding box [xmin, xmax, ymin, ymax]
    - alternative box [xmin, ymin, xmax, ymax]
    - centre and width+height
    - bottom-left and top-right corners
    - bottom-left corner and width+height
    - top-right corner and width+height
    - top-left corner and width+height

    For plots where the y-axis is inverted (eg. for images) then top is the
    smaller vertical coordinate.

    Example:

    .. runblock:: pycon

        >>> from spatialmath.base import plotvol2, plot_box
        >>> plotvol2(5)
        >>> plot_box('r', centre=(2,3), wh=1) # w=h=1
        >>> plot_box(tl=(1,1), br=(0,2), filled=True, color='b')
    """

    if wh is not None:
        if smbase.isscalar(wh):
            w, h = wh, wh
        else:
            w, h = wh
    
    # test for various 4-coordinate versions
    if bbox is not None:
        lb = bbox[:2]
        w, h = bbox[2:]

    elif lbwh is not None:
        lb = lbwh[:2]
        w, h = lbwh[2:]

    elif lbrt is not None:
        lb = lbrt[:2]
        rt = lbrt[2:]
        w, h = rt[0] - lb[0], rt[1] - lb[1]

    elif lrbt is not None:
        lb = (lrbt[0], lrbt[2])
        rt = (lrbt[1], lrbt[3])
        w, h = rt[0] - lb[0], rt[1] - lb[1]

    elif ltrb is not None:
        lb = (ltrb[0], ltrb[3])
        rt = (ltrb[2], ltrb[1])
        w, h = rt[0] - lb[0], rt[1] - lb[1]

    elif w is not None and h is not None:
        # we have width & height, one corner is enough

        if centre is not None:
            lb = (centre[0] - w/2, centre[1] - h/2)

        elif lt is not None:
            lb = (lt[0], lt[1] - h)

        elif rt is not None:
            lb = (rt[0] - w, rt[1] - h)

        elif rb is not None:
            lb = (rb[0] - w, rb[1])

    else:
        # we need two opposite corners
        if lb is not None and rt is not None:
            w = rt[0] - lb[0]
            h = rt[1] - lb[1]

        elif lt is not None and rb is not None:
            lb = (lt[0], rb[1])
            w = rb[0] - lt[0]
            h = lt[1] - rb[1]

        else:
            raise ValueError('cant compute box')

    if w < 0:
        raise ValueError("width must be positive")
    if h < 0:
        raise ValueError("height must be positive")

    # we only need lb, wh
    ax = axes_logic(ax, 2)
    if filled:
        r = plt.Rectangle(lb, w, h, clip_on=True, **kwargs)
    else:
        if 'color' in kwargs:
            kwargs['edgecolor'] = kwargs['color']
            del kwargs['color']
        r = plt.Rectangle(lb, w, h, clip_on=True, facecolor='None', **kwargs)
    ax.add_patch(r)

    return r

def plot_arrow(start, end, ax=None, **kwargs):
    """
    Plot 2D arrow

    :param start: start point, arrow tail
    :type start: array_like(2)
    :param end: end point, arrow head
    :type end: array_like(2)
    :param ax: axes to draw into, defaults to None
    :type ax: Axes, optional
    :param kwargs: argumetns to pass to :class:`matplotlib.patches.Arrow`

    Example:

    .. runblock:: pycon

        >>> from spatialmath.base import plotvol2, plot_arrow
        >>> plotvol2(5)
        >>> plot_arrow((-2, 2), (3, 4), color='r', width=0.1)  # red arrow
    """
    ax = axes_logic(ax, 2)

    ax.arrow(start[0], start[1], end[0] - start[0], end[1] - start[1], length_includes_head=True, **kwargs)

def plot_polygon(vertices, *fmt, close=False, **kwargs):
    """
    Plot polygon

    :param vertices: vertices
    :type vertices: ndarray(2,N)
    :param close: close the polygon, defaults to False
    :type close: bool, optional
    :param kwargs: arguments passed to Patch
    :return: Matplotlib artist
    :rtype: line or patch

    Example:

    .. runblock:: pycon

        >>> from spatialmath.base import plotvol2, plot_polygon
        >>> plotvol2(5)
        >>> vertices = np.array([[-1, 2, -1], [1, 0, -1]])
        >>> plot_polygon(vertices, filled=True, facecolor='g')  # green filled triangle
    """

    if close:
        vertices = np.hstack((vertices, vertices[:, [0]]))
    return _render2D(vertices, fmt=fmt, **kwargs)


def _render2D(vertices, pose=None, filled=False, color=None, ax=None, fmt=(), **kwargs):

    ax = axes_logic(ax, 2)
    if pose is not None:
        vertices = pose * vertices

    if filled:
        if color is not None:
            kwargs['facecolor'] = color
            kwargs['edgecolor'] = color
        r = plt.Polygon(vertices.T, closed=True, **kwargs)
        ax.add_patch(r)
    else:
        r = plt.plot(vertices[0, :], vertices[1, :], *fmt, color=color, **kwargs)
    return r


def circle(centre=(0, 0), radius=1, resolution=50, closed=False):
    """
    Points on a circle

    :param centre: centre of circle, defaults to (0, 0)
    :type centre: array_like(2), optional
    :param radius: radius of circle, defaults to 1
    :type radius: float, optional
    :param resolution: number of points on circumferece, defaults to 50
    :type resolution: int, optional
    :return: points on circumference
    :rtype: ndarray(2,N) or ndarray(3,N)

    Returns a set of ``resolution`` that lie on the circumference of a circle
    of given ``center`` and ``radius``.

    If ``len(centre)==3`` then the 3D coordinates are returned, where the
    circle lies in the xy-plane and the z-coordinate comes from ``centre[2]``.
    """
    if closed:
        resolution += 1
    u = np.linspace(0.0, 2.0 * np.pi, resolution, endpoint=closed)
    x = radius * np.cos(u) + centre[0]
    y = radius * np.sin(u) + centre[1]
    if len(centre) == 3:
        z = np.full(x.shape, centre[2])
        return np.array((x, y, z))
    else:
        return np.array((x, y))


def plot_circle(
    radius, centre, *fmt, resolution=50, ax=None, filled=False, **kwargs
):
    """
    Plot a circle using matplotlib

    :param centre: centre of circle, defaults to (0,0)
    :type centre: array_like(2), optional
    :param args:
    :param radius: radius of circle
    :type radius: float
    :param resolution: number of points on circumference, defaults to 50
    :type resolution: int, optional
    :return: the matplotlib object
    :rtype: list of Line2D or Patch.Polygon

    Plot or more circles. If ``centre`` is a 3xN array, then each column is
    taken as the centre of a circle.  All circles have the same radius, color
    etc.

    Example:

    .. runblock:: pycon

        >>> from spatialmath.base import plotvol2, plot_circle
        >>> plotvol2(5)
        >>> plot_circle(1, 'r')  # red circle
        >>> plot_circle(2, 'b--')  # blue dashed circle
        >>> plot_circle(0.5, filled=True, facecolor='y')  # yellow filled circle
    """
    centres = smbase.getmatrix(centre, (2, None))

    ax = axes_logic(ax, 2)
    handles = []
    for centre in centres.T:
        xy = circle(centre, radius, resolution, closed=not filled)
        if filled:
            patch = plt.Polygon(xy.T, **kwargs)
            handles.append(ax.add_patch(patch))
        else:
            handles.append(ax.plot(xy[0, :], xy[1, :], *fmt, **kwargs))
    return handles


def ellipse(E, centre=(0, 0), scale=1, confidence=None, resolution=40, inverted=False, closed=False):
    r"""
    Points on ellipse

    :param E: ellipse
    :type E: ndarray(2,2)
    :param centre: ellipse centre, defaults to (0,0,0)
    :type centre: tuple, optional
    :param scale: scale factor for the ellipse radii
    :type scale: float
    :param confidence: if E is an inverse covariance matrix plot an ellipse
        for this confidence interval in the range [0,1], defaults to None
    :type confidence: float, optional
    :param resolution: number of points on circumferance, defaults to 40
    :type resolution: int, optional
    :param inverted: if :math:`\mat{E}^{-1}` is provided, defaults to False
    :type inverted: bool, optional
    :raises ValueError: [description]
    :return: points on circumference
    :rtype: ndarray(2,N)

    The ellipse is defined by :math:`x^T \mat{E} x = s^2` where :math:`x \in
    \mathbb{R}^2` and :math:`s` is the scale factor.

    .. note:: For some common cases we require :math:`\mat{E}^{-1}`, for example
        - for robot manipulability
          :math:`\nu (\mat{J} \mat{J}^T)^{-1} \nu` i
        - a covariance matrix
          :math:`(x - \mu)^T \mat{P}^{-1} (x - \mu)`
        so to avoid inverting ``E`` twice to compute the ellipse, we flag that
        the inverse is provided using ``inverted``.
    """
    if E.shape != (2, 2):
        raise ValueError("ellipse is defined by a 2x2 matrix")

    if confidence:
        from scipy.stats.distributions import chi2

        # process the probability
        s = math.sqrt(chi2.ppf(confidence, df=2)) * scale
    else:
        s = scale

    xy = circle(resolution=resolution, closed=closed)  # unit circle

    if not inverted:
        E = np.linalg.inv(E)

    e = s * sp.linalg.sqrtm(E) @ xy + np.array(centre, ndmin=2).T
    return e


def plot_ellipse(
    E,
    *fmt,
    centre=(0, 0),
    scale=1,
    confidence=None,
    resolution=40,
    inverted=False,
    ax=None,
    filled=False,
    **kwargs
):
    r"""
    Plot an ellipse using matplotlib

    :param E: matrix describing ellipse
    :type E: ndarray(2,2)
    :param centre: centre of ellipse, defaults to (0, 0)
    :type centre: array_like(2), optional
    :param scale: scale factor for the ellipse radii
    :type scale: float
    :param resolution: number of points on circumferece, defaults to 40
    :type resolution: int, optional
    :return: the matplotlib object
    :rtype: Line2D or Patch.Polygon

    The ellipse is defined by :math:`x^T \mat{E} x = s^2` where :math:`x \in
    \mathbb{R}^2` and :math:`s` is the scale factor.

    .. note:: For some common cases we require :math:`\mat{E}^{-1}`, for example
        - for robot manipulability
          :math:`\nu (\mat{J} \mat{J}^T)^{-1} \nu` i
        - a covariance matrix
          :math:`(x - \mu)^T \mat{P}^{-1} (x - \mu)`
        so to avoid inverting ``E`` twice to compute the ellipse, we flag that
        the inverse is provided using ``inverted``.

    Returns a set of ``resolution``  that lie on the circumference of a circle
    of given ``center`` and ``radius``.

    Example:

    .. runblock:: pycon

        >>> from spatialmath.base import plotvol2, plot_circle
        >>> plotvol2(5)
        >>> plot_ellipse(np.diag((1,2)), 'r')  # red ellipse
        >>> plot_ellipse(np.diag((1,2)), 'b--')  # blue dashed ellipse
        >>> plot_ellipse(np.diag((1,2)), filled=True, facecolor='y')  # yellow filled ellipse

    """
    # allow for centre[2] to plot ellipse in a plane in a 3D plot

    xy = ellipse(E, centre, scale, confidence, resolution, inverted, closed=True)
    ax = axes_logic(ax, 2)
    if filled:
        patch = plt.Polygon(xy.T, **kwargs)
        ax.add_patch(patch)
    else:
        plt.plot(xy[0, :], xy[1, :], *fmt, **kwargs)


# =========================== 3D shapes =================================== #


def sphere(radius=1, centre=(0, 0, 0), resolution=50):
    """
    Points on a sphere

    :param centre: centre of sphere, defaults to (0, 0, 0)
    :type centre: array_like(3), optional
    :param radius: radius of sphere, defaults to 1
    :type radius: float, optional
    :param resolution: number of points ``N`` on circumferece, defaults to 50
    :type resolution: int, optional
    :return: X, Y and Z braid matrices
    :rtype: 3 x ndarray(N, N)

    :seealso: :func:`plot_sphere`, :func:`~matplotlib.pyplot.plot_surface`, :func:`~matplotlib.pyplot.plot_wireframe`
    """
    theta_range = np.linspace(0, np.pi, resolution)
    phi_range = np.linspace(-np.pi, np.pi, resolution)

    Phi, Theta = np.meshgrid(phi_range, theta_range)

    x = radius * np.sin(Theta) * np.cos(Phi) + centre[0]
    y = radius * np.sin(Theta) * np.sin(Phi) + centre[1]
    z = radius * np.cos(Theta) + centre[2]

    return (x, y, z)


def plot_sphere(radius, centre=(0, 0, 0), pose=None, resolution=50, ax=None, **kwargs):
    """
    Plot a sphere using matplotlib

    :param centre: centre of sphere, defaults to (0, 0, 0)
    :type centre: array_like(3), ndarray(3,N), optional
    :param radius: radius of sphere, defaults to 1
    :type radius: float, optional
    :param resolution: number of points on circumferece, defaults to 50
    :type resolution: int, optional

    :param pose: pose of sphere, defaults to None
    :type pose: SE3, optional
    :param ax: axes to draw into, defaults to None
    :type ax: Axes3D, optional
    :param filled: draw filled polygon, else wireframe, defaults to False
    :type filled: bool, optional
    :param kwargs: arguments passed to ``plot_wireframe`` or ``plot_surface``

    :return: matplotlib collection
    :rtype: list of Line3DCollection or Poly3DCollection

    Plot one or more spheres. If ``centre`` is a 3xN array, then each column is
    taken as the centre of a sphere.  All spheres have the same radius, color
    etc.

    Example:

    .. runblock:: pycon

        >>> from spatialmath.base import plot_sphere
        >>> plot_sphere(radius=1, color='r')   # red sphere wireframe
        >>> plot_sphere(radius=1, centre=(1,1,1), filled=True, facecolor='b')

    :seealso: :func:`~matplotlib.pyplot.plot_surface`, :func:`~matplotlib.pyplot.plot_wireframe`
    """
    ax = axes_logic(ax, 3)

    centre = smbase.getmatrix(centre, (3, None))

    handles = []
    for c in centre.T:
        X, Y, Z = sphere(centre=c, radius=radius, resolution=resolution)
        handles.append(_render3D(ax, X, Y, Z, **kwargs))

    return handles


def ellipsoid(
    E, centre=(0, 0, 0), scale=1, confidence=None, resolution=40, inverted=False
):
    r"""
    rPoints on an ellipsoid

    :param centre: centre of ellipsoid, defaults to (0, 0, 0)
    :type centre: array_like(3), optional
    :param scale: scale factor for the ellipse radii
    :type scale: float
    :param confidence: confidence interval, range 0 to 1
    :type confidence: float
    :param resolution: number of points ``N`` on circumferece, defaults to 40
    :type resolution: int, optional
    :param inverted: :math:`E^{-1}` rather than :math:`E` provided, defaults to False
    :type inverted: bool, optional
    :return: X, Y and Z braid matrices
    :rtype: 3 x ndarray(N, N)

    The ellipse is defined by :math:`x^T \mat{E} x = s^2` where :math:`x \in
    \mathbb{R}^3` and :math:`s` is the scale factor.

    .. note:: For some common cases we require :math:`\mat{E}^{-1}`, for example
        - for robot manipulability
          :math:`\nu (\mat{J} \mat{J}^T)^{-1} \nu` i
        - a covariance matrix
          :math:`(x - \mu)^T \mat{P}^{-1} (x - \mu)`
        so to avoid inverting ``E`` twice to compute the ellipse, we flag that
        the inverse is provided using ``inverted``.

    :seealso: :func:`plot_ellipsoid`, :func:`~matplotlib.pyplot.plot_surface`, :func:`~matplotlib.pyplot.plot_wireframe`
    """
    if E.shape != (3, 3):
        raise ValueError("ellipsoid is defined by a 3x3 matrix")

    if confidence:
        # process the probability
        from scipy.stats.distributions import chi2

        s = math.sqrt(chi2.ppf(confidence, df=3)) * scale
    else:
        s = scale

    if not inverted:
        E = np.linalg.inv(E)

    x, y, z = sphere()  # unit sphere
    e = (
        s * sp.linalg.sqrtm(E) @ np.array([x.flatten(), y.flatten(), z.flatten()])
        + np.c_[centre].T
    )
    return e[0, :].reshape(x.shape), e[1, :].reshape(x.shape), e[2, :].reshape(x.shape)


def plot_ellipsoid(
    E,
    centre=(0, 0, 0),
    scale=1,
    confidence=None,
    resolution=40,
    inverted=False,
    ax=None,
    **kwargs
):
    r"""
    Draw an ellipsoid using matplotlib

    :param E: ellipsoid
    :type E: ndarray(3,3)
    :param centre: [description], defaults to (0,0,0)
    :type centre: tuple, optional
    :param scale:
    :type scale:
    :param confidence: confidence interval, range 0 to 1
    :type confidence: float
    :param resolution: number of points on circumferece, defaults to 40
    :type resolution: int, optional
    :param inverted: :math:`E^{-1}` rather than :math:`E` provided, defaults to False
    :type inverted: bool, optional
    :param ax: [description], defaults to None
    :type ax: [type], optional
    :param wireframe: [description], defaults to False
    :type wireframe: bool, optional
    :param stride: [description], defaults to 1
    :type stride: int, optional

    ``plot_ellipse(E)`` draws the ellipsoid defined by :math:`x^T \mat{E} x = 0`
    on the current plot.

    Example::

          H = plot_ellipse(diag([1 2]), [3 4]', 'r'); % draw red ellipse
          plot_ellipse(diag([1 2]), [5 6]', 'alter', H); % move the ellipse
          plot_ellipse(diag([1 2]), [5 6]', 'alter', H, 'LineColor', 'k'); % change color

          plot_ellipse(COVAR, 'confidence', 0.95); % draw 95% confidence ellipse

    .. note::

        - If a confidence interval is given then ``E`` is interpretted as a covariance
          matrix and the ellipse size is computed using an inverse chi-squared function.

    :seealso: :func:`~matplotlib.pyplot.plot_surface`, :func:`~matplotlib.pyplot.plot_wireframe`
    """
    X, Y, Z = ellipsoid(E, centre, scale, confidence, resolution, inverted)
    ax = axes_logic(ax, 3)
    handle = _render3D(ax, X, Y, Z, **kwargs)
    return [handle]

# TODO, get cylinder, cuboid, cone working
def cylinder(center_x, center_y, radius, height_z, resolution=50):
    Z = np.linspace(0, height_z, radius)
    theta = np.linspace(0, 2 * np.pi, radius)
    theta_grid, z_grid = np.meshgrid(theta, z)
    X = radius * np.cos(theta_grid) + center_x
    Y = radius * np.sin(theta_grid) + center_y
    return X, Y, Z

# https://stackoverflow.com/questions/30715083/python-plotting-a-wireframe-3d-cuboid
# https://stackoverflow.com/questions/26874791/disconnected-surfaces-when-plotting-cones
def plot_cylinder(
    radius,
    height,
    resolution=50,
    centre=(0, 0, 0),
    ends=False,
    ax=None,
    filled=False,
    **kwargs
):
    """
    Plot a cylinder using matplotlib

    :param radius: radius of sphere
    :type radius: float
    :param height: height of cylinder in the z-direction
    :type height: float or array_like(2)
    :param resolution: number of points on circumferece, defaults to 50
    :type resolution: int, optional

    :param pose: pose of sphere, defaults to None
    :type pose: SE3, optional
    :param ax: axes to draw into, defaults to None
    :type ax: Axes3D, optional
    :param filled: draw filled polygon, else wireframe, defaults to False
    :type filled: bool, optional
    :param kwargs: arguments passed to ``plot_wireframe`` or ``plot_surface``

    :return: matplotlib objects
    :rtype: list of matplotlib object types

    The axis of the cylinder is parallel to the z-axis and extends from z=0
    to z=height, or z=height[0] to z=height[1].

    The cylinder can be positioned by setting ``centre``, or positioned
    and orientated by setting ``pose``.

    :seealso: :func:`~matplotlib.pyplot.plot_surface`, :func:`~matplotlib.pyplot.plot_wireframe`
    """
    if smbase.isscalar(height):
        height = [0, height]

    ax = axes_logic(ax, 3)
    x = np.linspace(centre[0] - radius, centre[0] + radius, resolution)
    z = height
    X, Z = np.meshgrid(x, z)

    Y = np.sqrt(radius ** 2 - (X - centre[0]) ** 2) + centre[1]  # Pythagorean theorem

    handles = []
    handles.append(_render3D(ax, X, Y, Z, filled=filled, **kwargs))
    handles.append(_render3D(ax, X, (2 * centre[1] - Y), Z, filled=filled, **kwargs))

    if ends and kwargs.get("filled", default=False):
        floor = Circle(centre[:2], radius, **kwargs)
        handles.append(ax.add_patch(floor))
        pathpatch_2d_to_3d(floor, z=height[0], zdir="z")

        ceiling = Circle(centre[:2], radius, **kwargs)
        handles.append(ax.add_patch(ceiling))
        pathpatch_2d_to_3d(ceiling, z=height[1], zdir="z")

    return handles


def plot_cone(
    radius,
    height,
    resolution=50,
    flip=False,
    centre=(0, 0, 0),
    ends=False,
    ax=None,
    filled=False,
    **kwargs
):
    """
    Plot a cone using matplotlib

    :param radius: radius of cone at open end
    :type radius: float
    :param height: height of cone in the z-direction
    :type height: float
    :param resolution: number of points on circumferece, defaults to 50
    :type resolution: int, optional
    :param flip: cone faces upward, defaults to False
    :type flip: bool, optional

    :param pose: pose of cone, defaults to None
    :type pose: SE3, optional
    :param ax: axes to draw into, defaults to None
    :type ax: Axes3D, optional
    :param filled: draw filled polygon, else wireframe, defaults to False
    :type filled: bool, optional
    :param kwargs: arguments passed to ``plot_wireframe`` or ``plot_surface``

    :return: matplotlib objects
    :rtype: list of matplotlib object types

    The axis of the cone is parallel to the z-axis and it is drawn pointing
    down. The point is at z=0 and the open end at z= ``height``.  If ``flip`` is
    True then the cone faces upwards, the point is at z= ``height`` and the open
    end at z=0.

    The cylinder can be positioned by setting ``centre``, or positioned
    and orientated by setting ``pose``.

    :seealso: :func:`~matplotlib.pyplot.plot_surface`, :func:`~matplotlib.pyplot.plot_wireframe`
    """
    ax = axes_logic(ax, 3)

    # https://stackoverflow.com/questions/26874791/disconnected-surfaces-when-plotting-cones
    # Set up the grid in polar coords
    theta = np.linspace(0, 2 * np.pi, resolution)
    r = np.linspace(0, radius, resolution)
    T, R = np.meshgrid(theta, r)

    # Then calculate X, Y, and Z
    X = R * np.cos(T) + centre[0]
    Y = R * np.sin(T) + centre[1]
    Z = np.sqrt(X ** 2 + Y ** 2) / radius * height + centre[2]
    if flip:
        Z = height - Z

    handles = []
    handles.append(_render3D(ax, X, Y, Z, filled=filled, **kwargs))
    handles.append(_render3D(ax, X, (2 * centre[1] - Y), Z, filled=filled, **kwargs))

    if ends and kwargs.get("filled", default=False):
        floor = Circle(centre[:2], radius, **kwargs)
        handles.append(ax.add_patch(floor))
        pathpatch_2d_to_3d(floor, z=height[0], zdir="z")

        ceiling = Circle(centre[:2], radius, **kwargs)
        handles.append(ax.add_patch(ceiling))
        pathpatch_2d_to_3d(ceiling, z=height[1], zdir="z")

    return handles


def plot_cuboid(
    sides=[1, 1, 1], centre=(0, 0, 0), pose=None, ax=None, filled=False, **kwargs
):
    """
    Plot a cuboid (3D box) using matplotlib

    :param sides: side lengths, defaults to 1
    :type sides: array_like(3), optional
    :param centre: centre of box, defaults to (0, 0, 0)
    :type centre: array_like(3), optional

    :param pose: pose of sphere, defaults to None
    :type pose: SE3, optional
    :param ax: axes to draw into, defaults to None
    :type ax: Axes3D, optional
    :param filled: draw filled polygon, else wireframe, defaults to False
    :type filled: bool, optional
    :param kwargs: arguments passed to ``plot_wireframe`` or ``plot_surface``

    :return: matplotlib collection
    :rtype: Line3DCollection or Poly3DCollection

    :seealso: :func:`~matplotlib.pyplot.plot_surface`, :func:`~matplotlib.pyplot.plot_wireframe`
    """

    vertices = (
        np.array(
            list(
                product(
                    [-sides[0], sides[0]], [-sides[1], sides[1]], [-sides[2], sides[2]]
                )
            )
        )
        / 2
        + centre
    )
    vertices = vertices.T

    if pose is not None:
        vertices = smbase.homtrans(pose.A, vertices)

    ax = axes_logic(ax, 3)
    # plot sides
    if filled:
        # show faces

        faces = [
            [0, 1, 3, 2],
            [4, 5, 7, 6],  # YZ planes
            [0, 1, 5, 4],
            [2, 3, 7, 6],  # XZ planes
            [0, 2, 6, 4],
            [1, 3, 7, 5],  # XY planes
        ]
        F = [[vertices[:, i] for i in face] for face in faces]
        collection = Poly3DCollection(F, **kwargs)
        ax.add_collection3d(collection)
        return collection
    else:
        edges = [[0, 1, 3, 2, 0], [4, 5, 7, 6, 4], [0, 4], [1, 5], [3, 7], [2, 6]]
        lines = []
        for edge in edges:
            E = vertices[:, edge]
            # ax.plot(E[0], E[1], E[2], **kwargs)
            lines.append(E.T)
        if 'color' in kwargs:
            if 'alpha' in kwargs:
                alpha = kwargs['alpha']
                del kwargs['alpha']
            else:
                alpha = 1
            kwargs['colors'] = colors.to_rgba(kwargs['color'], alpha)
            del kwargs['color']
        collection = Line3DCollection(lines, **kwargs)
        ax.add_collection3d(collection)
        return collection


def _render3D(ax, X, Y, Z, pose=None, filled=False, color=None, **kwargs):

    # TODO:
    # handle pose in here
    # do the guts of plot_surface/wireframe but skip the auto scaling
    # have all 3d functions use this
    # rename all functions with 3d suffix sphere3d, box3d, ell

    if pose is not None:
        # long version:
        # xc = X.reshape((-1,))
        # yc = Y.reshape((-1,))
        # zc = Z.reshape((-1,))
        # xyz = np.array((xc, yc, zc))
        # xyz = pose * xyz
        # X = xyz[0, :].reshape(X.shape)
        # Y = xyz[1, :].reshape(Y.shape)
        # Z = xyz[2, :].reshape(Z.shape)

        # short version:
        xyz = pose * np.dstack((X, Y, Z)).reshape((-1, 3)).T
        X, Y, Z = np.squeeze(np.dsplit(xyz.T.reshape(X.shape + (3,)), 3))

    if filled:
        return ax.plot_surface(X, Y, Z, color=color, **kwargs)
    else:
        kwargs["colors"] = color
        return ax.plot_wireframe(X, Y, Z, **kwargs)


def _axes_dimensions(ax):
    """
    Dimensions of axes

    :param ax: axes
    :type ax: Axes3DSubplot or AxesSubplot
    :return: dimensionality of axes, either 2 or 3
    :rtype: int
    """
    classname = ax.__class__.__name__

    if classname in ("Axes3DSubplot", "Animate"):
        return 3
    elif classname in ("AxesSubplot", "Animate2"):
        return 2


def axes_get_limits(ax):
    return np.r_[ax.get_xlim(), ax.get_ylim()]


def axes_get_scale(ax):
    limits = axes_get_limits(ax)
    return max(abs(limits[1] - limits[0]), abs(limits[3] - limits[2]))


def axes_logic(ax, dimensions, projection="ortho", autoscale=True):
    """
    Axis creation logic

    :param ax: axes to draw in
    :type ax: Axes3DSubplot, AxesSubplot or None
    :param dimensions: required dimensionality, 2 or 3
    :type dimensions: int
    :param projection: 3D projection type, defaults to 'ortho'
    :type projection: str, optional
    :return: axes to draw in
    :rtype: Axes3DSubplot or AxesSubplot

    Given a request for axes with either 2 or 3 dimensions it checks for a
    match with the passed axes ``ax`` or the current axes.

    If the dimensions do not match, or no figure/axes currently exist,
    then ``plt.axes()`` is called to create one.

    Used by all plot_xxx() functions in this module.
    """

    if not _matplotlib_exists:
        raise NotImplementedError("Matplotlib is not installed: pip install matplotlib")

    # print(f"new axis logic ({dimensions}D): ", end='')
    if ax is None:
        # no axes passed in, find out what's happening
        # need to be careful to not use gcf() or gca() since they
        # auto create fig/axes if none exist
        nfigs = len(plt.get_fignums())
        if nfigs > 0:
            # there are figures
            fig = plt.gcf()  # get current figure
            naxes = len(fig.axes)
            # print(f"existing fig with {naxes} axes")
            if naxes > 0:
                ax = plt.gca()  # get current axes
                if _axes_dimensions(ax) == dimensions:
                    return ax
        # otherwise it doesnt exist or dimension mismatch, create new axes

    else:
        # axis was given

        if _axes_dimensions(ax) == dimensions:
            # print("use existing axes")
            return ax
        # mismatch in dimensions, create new axes
    # print('create new axes')
    plt.figure()
    # no axis specified
    if dimensions == 2:
        ax = plt.axes()
        if autoscale:
            ax.autoscale()
    else:
        ax = plt.axes(projection="3d", proj_type=projection)

    plt.sca(ax)
    plt.axes(ax)

    return ax


def plotvol2(dim, ax=None, equal=True, grid=False, labels=True):
    """
    Create 2D plot area

    :param ax: axes of initializer, defaults to new subplot
    :type ax: AxesSubplot, optional
    :param equal: set aspect ratio to 1:1, default False
    :type equal: bool
    :return: initialized axes
    :rtype: AxesSubplot

    Initialize axes with dimensions given by ``dim`` which can be:

    ==============  ======  ======
    input           xrange  yrange
    ==============  ======  ======
    A (scalar)      -A:A    -A:A
    [A, B]           A:B     A:B
    [A, B, C, D]     A:B     C:D
    ==============  ======  ======

    :seealso: :func:`plotvol3`, :func:`expand_dims`
    """
    dims = expand_dims(dim, 2)
    if ax is None:
        ax = plt.subplot()
    ax.axis(dims)
    if labels:
        ax.set_xlabel("X")
        ax.set_ylabel("Y")

    if equal:
        ax.set_aspect("equal")
    if grid:
        ax.grid(True)
        ax.set_axisbelow(True)

    # signal to related functions that plotvol set the axis limits
    ax._plotvol = True
    return ax


def plotvol3(
    dim=None, ax=None, equal=True, grid=False, labels=True, projection="ortho"
):
    """
    Create 3D plot volume

    :param ax: axes of initializer, defaults to new subplot
    :type ax: Axes3DSubplot, optional
    :param equal: set aspect ratio to 1:1:1, default False
    :type equal: bool
    :return: initialized axes
    :rtype: Axes3DSubplot

    Initialize axes with dimensions given by ``dim`` which can be:

    ==================  ======  ======  =======
    input               xrange  yrange  zrange
    ==================  ======  ======  =======
    A (scalar)          -A:A    -A:A    -A:A
    [A, B]              A:B     A:B     A:B
    [A, B, C, D, E, F]  A:B     C:D     E:F
    ==================  ======  ======  =======

    :seealso: :func:`plotvol2`, :func:`expand_dims`
    """
    # create an axis if none existing
    ax = axes_logic(ax, 3, projection=projection)

    if dim is None:
        ax.autoscale(True)
    else:
        dims = expand_dims(dim, 3)
        ax.set_xlim3d(dims[0], dims[1])
        ax.set_ylim3d(dims[2], dims[3])
        ax.set_zlim3d(dims[4], dims[5])
        if labels:
            ax.set_xlabel("X")
            ax.set_ylabel("Y")
            ax.set_zlabel("Z")

    if equal:
        try:
            ax.set_box_aspect((1,) * 3)
        except AttributeError:
            # old version of MPL doesn't support this
            warnings.warn(
                "Current version of matplotlib does not support set_box_aspect()"
            )
    if grid:
        ax.grid(True)

    # signal to related functions that plotvol set the axis limits
    ax._plotvol = True
    return ax


def expand_dims(dim=None, nd=2):
    """
    Expand compact axis dimensions

    :param dim: dimensions, defaults to None
    :type dim: scalar, array_like(2), array_like(4), array_like(6), optional
    :param nd: number of axes dimensions, defaults to 2
    :type nd: int, optional
    :raises ValueError: bad arguments
    :return: 2d or 3d dimensions vector
    :rtype: ndarray(4) or ndarray(6)

    Compute bounding dimensions for plots from shorthand notation.

    If ``nd==2``, [xmin, xmax, ymin, ymax]:
        * A -> [-A, A, -A, A]
        * [A,B] -> [A, B, A, B]
        * [A,B,C,D] -> [A, B, C, D]

    If ``nd==3``, [xmin, xmax, ymin, ymax, zmin, zmax]:
        * A -> [-A, A, -A, A, -A, A]
        * [A,B] -> [A, B, A, B, A, B]
        * [A,B,C,D,E,F] -> [A, B, C, D, E, F]
    """
    dim = smbase.getvector(dim)

    if nd == 2:
        if len(dim) == 1:
            return np.r_[-dim, dim, -dim, dim]
        elif len(dim) == 2:
            return np.r_[dim[0], dim[1], dim[0], dim[1]]
        elif len(dim) == 4:
            return dim
        else:
            raise ValueError("bad dimension specified")
    elif nd == 3:
        if len(dim) == 1:
            return np.r_[-dim, dim, -dim, dim, -dim, dim]
        elif len(dim) == 2:
            return np.r_[dim[0], dim[1], dim[0], dim[1], dim[0], dim[1]]
        elif len(dim) == 6:
            return dim
        else:
            raise ValueError("bad dimension specified")
    else:
        raise ValueError("nd is 2 or 3")


def isnotebook():
    """
    Determine if code is being run from a Jupyter notebook

    :references:

        - https://stackoverflow.com/questions/15411967/how-can-i-check-if-code-
          is-executed-in-the-ipython-notebook/39662359#39662359
    """
    try:
        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False  # Probably standard Python interpreter


if __name__ == "__main__":
    import pathlib


    plotvol2(5)
    # plot_box(ltrb=[-1, 4, 2, 2], color='r', linewidth=2)
    # plot_box(lbrt=[-1, 2, 2, 4], color='k', linestyle='--', linewidth=4)
    # plot_box(lbwh=[2, -2, 2, 3], color='k', linewidth=2)
    # plot_box(centre=(-2, -1), wh=2, color='b', linewidth=2)
    # plot_box(centre=(-2, -1), wh=(1,3), color='g', linewidth=2)
    # plt.grid(True)
    # plt.show(block=True)

    # plt.imshow(np.eye(200))
    # umin, umax, vmin, vmax = 23, 166, 110, 212
    # plot_box(l=umin, r=umax, t=vmin, b=vmax, color="g")

    plot_circle(1, (2,3), resolution=3, filled=False)
    plt.show(block=True)
    
    exec(
        open(
            pathlib.Path(__file__).parent.parent.parent.absolute()
            / "tests"
            / "base"
            / "test_graphics.py"
        ).read()
    )  # pylint: disable=exec-used



import numpy.testing as nt
import matplotlib.pyplot as plt
import unittest

"""
we will assume that the primitives rotx,trotx, etc. all work
"""
from math import pi
from spatialmath.twist import *
# from spatialmath import super_pose # as sp
from spatialmath.base import *
from spatialmath.base import argcheck
from spatialmath.baseposematrix import BasePoseMatrix
from spatialmath.twist import BaseTwist

def array_compare(x, y):
    if isinstance(x, BasePoseMatrix):
        x = x.A
    if isinstance(y, BasePoseMatrix):
        y = y.A
    if isinstance(x, BaseTwist):
        x = x.S
    if isinstance(y, BaseTwist):
        y = y.S
    nt.assert_array_almost_equal(x, y)


class Twist3dTest(unittest.TestCase):
    
    def test_constructor(self):
        
        s = [1, 2, 3, 4, 5, 6]
        x = Twist3(s)
        self.assertIsInstance(x, Twist3)
        self.assertEqual(len(x), 1)
        array_compare(x.v, [1, 2, 3])
        array_compare(x.w, [4, 5, 6])
        array_compare(x.S, s)
        
        x = Twist3([1,2,3], [4,5,6])
        array_compare(x.v, [1, 2, 3])
        array_compare(x.w, [4, 5, 6])
        array_compare(x.S, s)

        y = Twist3(x)
        array_compare(x, y)
        
        x = Twist3(SE3())
        array_compare(x, [0,0,0,0,0,0])
        
        
    def test_list(self):
        x = Twist3([1, 0, 0, 0, 0, 0])
        y = Twist3([1, 0, 0, 0, 0, 0])
                   
        a = Twist3(x)
        a.append(y)
        self.assertEqual(len(a), 2)
        array_compare(a[0], x)
        array_compare(a[1], y)
        
    def test_conversion_SE3(self):
        T = SE3.Rx(0)
        tw = Twist3(T)
        array_compare(tw.SE3(), T)
        self.assertIsInstance(tw.SE3(), SE3)
        self.assertEqual(len(tw.SE3()), 1)

        T = SE3.Rx(0) * SE3(1, 2, 3)
        array_compare(Twist3(T).SE3(), T)
        
    def test_conversion_se3(self):
        s = [1, 2, 3, 4, 5, 6]
        x = Twist3(s)
        
        array_compare(x.skewa(), np.array([[ 0., -6.,  5.,  1.],
                    [ 6.,  0., -4.,  2.],
                    [-5.,  4.,  0.,  3.],
                    [ 0.,  0.,  0.,  0.]]))
        
    def test_conversion_Plucker(self):
        pass
        
    def test_list_constuctor(self):
        x = Twist3([1, 0, 0, 0, 0, 0])
        
        a = Twist3([x,x,x,x])
        self.assertIsInstance(a, Twist3)
        self.assertEqual(len(a), 4)
        
        a = Twist3([x.skewa(), x.skewa(), x.skewa(), x.skewa()])
        self.assertIsInstance(a, Twist3)
        self.assertEqual(len(a), 4)
        
        a = Twist3([x.S, x.S, x.S, x.S])
        self.assertIsInstance(a, Twist3)
        self.assertEqual(len(a), 4)
        
        s = np.r_[1, 2, 3, 4, 5, 6]
        a = Twist3([s, s, s, s])
        self.assertIsInstance(a, Twist3)
        self.assertEqual(len(a), 4)
        
    def test_predicate(self):
        x = Twist3.UnitRevolute([1, 2, 3], [0, 0, 0])
        self.assertFalse(x.isprismatic)
        
        # check prismatic twist
        x = Twist3.UnitPrismatic([1, 2, 3])
        self.assertTrue(x.isprismatic)
        
        self.assertTrue(Twist3.isvalid(x.skewa()))
        self.assertTrue(Twist3.isvalid(x.S))
        
        self.assertFalse(Twist3.isvalid(2))
        self.assertFalse(Twist3.isvalid(np.eye(4)))
        
    def test_str(self):
        x = Twist3([1, 2, 3, 4, 5, 6])
        s = str(x)
        self.assertIsInstance(s, str)
        self.assertEqual(len(s), 14)
        self.assertEqual(s.count('\n'), 0)
        
        x.append(x)
        s = str(x)
        self.assertIsInstance(s, str)
        self.assertEqual(len(s), 29)
        self.assertEqual(s.count('\n'), 1)
        
    def test_variant_constructors(self):
        
        # check rotational twist
        x = Twist3.UnitRevolute([1, 2, 3], [0, 0, 0])
        array_compare(x, np.r_[0, 0, 0, unitvec([1, 2, 3])])
        
        # check prismatic twist
        x = Twist3.UnitPrismatic([1, 2, 3])
        array_compare(x, np.r_[unitvec([1, 2, 3]), 0, 0, 0, ])
    
    def test_SE3_twists(self):
        tw = Twist3( SE3.Rx(0) )
        array_compare(tw, np.r_[0, 0, 0,  0, 0, 0])
                      
        tw = Twist3( SE3.Rx(pi / 2) )
        array_compare(tw, np.r_[0, 0, 0,  pi / 2, 0, 0])
                      
        tw = Twist3( SE3.Ry(pi / 2) )
        array_compare(tw, np.r_[0, 0, 0,  0, pi / 2, 0])
                      
        tw = Twist3( SE3.Rz(pi / 2) )
        array_compare(tw, np.r_[0, 0, 0,  0, 0, pi / 2])
        
        tw = Twist3( SE3([1, 2, 3]) )
        array_compare(tw, [1, 2, 3,  0, 0, 0])
                      
        tw = Twist3( SE3([1, 2, 3]) * SE3.Ry(pi / 2))
        array_compare(tw, np.r_[-pi / 2, 2, pi,  0, pi / 2, 0])
        
    def test_exp(self):
        tw = Twist3.UnitRevolute([1, 0, 0], [0, 0, 0])
        array_compare(tw.exp(pi/2), SE3.Rx(pi/2))
        
        tw = Twist3.UnitRevolute([0, 1, 0], [0, 0, 0])
        array_compare(tw.exp(pi/2), SE3.Ry(pi/2))
        
        tw = Twist3.UnitRevolute([0, 0, 1], [0, 0, 0])
        array_compare(tw.exp(pi/2), SE3.Rz(pi / 2))
    
    def test_arith(self):
        
        # check overloaded *

        T1 = SE3(1, 2, 3) * SE3.Rx(pi / 2)
        T2 = SE3(4, 5, -6) * SE3.Ry(-pi / 2)
        
        x1 = Twist3(T1)
        x2 = Twist3(T2)

        array_compare( (x1 * x2).exp(), T1 * T2)
        array_compare( (x2 * x1).exp(), T2 * T1)
        
    def test_prod(self):
        # check prod
        T1 = SE3(1, 2, 3) * SE3.Rx(pi / 2)
        T2 = SE3(4, 5, -6) * SE3.Ry(-pi / 2)
        
        x1 = Twist3(T1)
        x2 = Twist3(T2)
        
        x = Twist3([x1, x2])
        array_compare( x.prod().SE3(), T1 * T2)
        

class Twist2dTest(unittest.TestCase):
    
    def test_constructor(self):
        
        s = [1, 2, 3]
        x = Twist2(s)
        self.assertIsInstance(x, Twist2)
        self.assertEqual(len(x), 1)
        array_compare(x.v, [1, 2])
        array_compare(x.w, [3])
        array_compare(x.S, s)
        
        x = Twist2([1,2], 3)
        array_compare(x.v, [1, 2])
        array_compare(x.w, [3])
        array_compare(x.S, s)

        y = Twist2(x)
        array_compare(x, y)
        
        # construct from SE2
        x = Twist2(SE2())
        array_compare(x, [0,0,0])
        
        x = Twist2( SE2(0, 0, pi / 2))
        array_compare(x, np.r_[0, 0, pi / 2])
        
        x = Twist2( SE2(1, 2,0 ))
        array_compare(x, np.r_[1, 2, 0])
        
        x = Twist2( SE2(1, 2, pi / 2))
        array_compare(x, np.r_[3 * pi / 4, pi / 4, pi / 2])
        
        
    def test_list(self):
        x = Twist2([1, 0, 0])
        y = Twist2([1, 0, 0])
                   
        a = Twist2(x)
        a.append(y)
        self.assertEqual(len(a), 2)
        array_compare(a[0], x)
        array_compare(a[1], y)

    def test_variant_constructors(self):
        
        # check rotational twist
        x = Twist2.UnitRevolute([1, 2])
        array_compare(x, np.r_[2, -1, 1])
        
        # check prismatic twist
        x = Twist2.UnitPrismatic([1, 2])
        array_compare(x, np.r_[unitvec([1, 2]), 0])
        
    def test_conversion_SE2(self):
        T = SE2(1, 2, 0.3)
        tw = Twist2(T)
        array_compare(tw.SE2(), T)
        self.assertIsInstance(tw.SE2(), SE2)
        self.assertEqual(len(tw.SE2()), 1)
        
    def test_conversion_se2(self):
        s = [1, 2, 3]
        x = Twist2(s)
        
        array_compare(x.skewa(), np.array([[ 0., -3.,  1.],
                                       [ 3.,  0.,  2.],
                                       [ 0.,  0.,  0.]]))

    def test_list_constuctor(self):
        x = Twist2([1, 0, 0])
        
        a = Twist2([x,x,x,x])
        self.assertIsInstance(a, Twist2)
        self.assertEqual(len(a), 4)
        
        a = Twist2([x.skewa(), x.skewa(), x.skewa(), x.skewa()])
        self.assertIsInstance(a, Twist2)
        self.assertEqual(len(a), 4)
        
        a = Twist2([x.S, x.S, x.S, x.S])
        self.assertIsInstance(a, Twist2)
        self.assertEqual(len(a), 4)
        
        s = np.r_[1, 2, 3]
        a = Twist2([s, s, s, s])
        self.assertIsInstance(a, Twist2)
        self.assertEqual(len(a), 4)
        
    def test_predicate(self):
        x = Twist2.UnitRevolute([1, 2])
        self.assertFalse(x.isprismatic)
        
        # check prismatic twist
        x = Twist2.UnitPrismatic([1, 2])
        self.assertTrue(x.isprismatic)
        
        self.assertTrue(Twist2.isvalid(x.skewa()))
        self.assertTrue(Twist2.isvalid(x.S))
        
        self.assertFalse(Twist2.isvalid(2))
        self.assertFalse(Twist2.isvalid(np.eye(3)))
        
    def test_str(self):
        x = Twist2([1, 2, 3])
        s = str(x)
        self.assertIsInstance(s, str)
        self.assertEqual(len(s), 8)
        self.assertEqual(s.count('\n'), 0)
        
        x.append(x)
        s = str(x)
        self.assertIsInstance(s, str)
        self.assertEqual(len(s), 17)
        self.assertEqual(s.count('\n'), 1)
        

    def test_SE2_twists(self):
        tw = Twist2( SE2() )
        array_compare(tw, np.r_[0, 0, 0])
                      
        tw = Twist2( SE2(0, 0, pi / 2) )
        array_compare(tw, np.r_[0, 0, pi / 2])
                      
        
        tw = Twist2( SE2([1, 2, 0]) )
        array_compare(tw, [1, 2, 0])
                      
        tw = Twist2( SE2([1, 2, pi / 2]))
        array_compare(tw, np.r_[ 3 * pi / 4, pi / 4, pi / 2])
        
    def test_exp(self):
        x = Twist2.UnitRevolute([0, 0])
        array_compare(x.exp(pi/2), SE2(0, 0, pi/2))
        
        x = Twist2.UnitRevolute([1, 0])
        array_compare(x.exp(pi/2), SE2(1, -1, pi/2))
        
        x = Twist2.UnitRevolute([1, 2])
        array_compare(x.exp(pi/2), SE2(3, 1, pi/2))

    
    def test_arith(self):
        
        # check overloaded *

        T1 = SE2(1, 2, pi / 2)
        T2 = SE2(4, 5, -pi / 4)
        
        x1 = Twist2(T1)
        x2 = Twist2(T2)

        array_compare( (x1 * x2).exp(), (T1 * T2).A)
        array_compare( (x2 * x1).exp(), (T2 * T1).A)

        array_compare( (x1 * x2).SE2(), (T1 * T2).A)
        array_compare( (x2 * x1).SE2(), (T2 * T1))
        
    def test_prod(self):
        # check prod
        T1 = SE2(1, 2, pi / 2)
        T2 = SE2(4, 5, -pi / 4)
        
        x1 = Twist2(T1)
        x2 = Twist2(T2)
        
        x = Twist2([x1, x2])
        array_compare( x.prod().SE2(), T1 * T2)

# ---------------------------------------------------------------------------------------#
if __name__ == '__main__':


    unittest.main()
